from typing import List

import re

import pandas as pd
from vacances_scolaires_france import SchoolHolidayDates


def lockdown_format_func(lockdown_idx: int) -> str:
    return f"Lockdown {lockdown_idx + 1}"


def get_school_holidays_FR(years: List[int]) -> pd.DataFrame:
    """Retrieve french school holidays and transform it into a Prophet holidays compatible df

    Parameters
    ----------
    years: List[int]
        List of years for which to retrieve holidays.

    Returns
    -------
    pd.DataFrame
        Holidays dataframe with columns 'ds' and 'holiday'.
    """

    def _get_school_holidays_FR_for_year(year: int) -> pd.DataFrame:
        fr_holidays = SchoolHolidayDates()
        df_vacances = pd.DataFrame.from_dict(fr_holidays.holidays_for_year(year)).T.reset_index(
            drop=True
        )
        df_vacances = df_vacances.rename(columns={"date": "ds", "nom_vacances": "holiday"})
        df_vacances["holiday"] = df_vacances["holiday"].apply(
            lambda x: re.sub(r"^Vacances (De|D')? ?(La )?", "School holiday: ", x.title())
        )
        df_vacances["ds"] = pd.to_datetime(df_vacances["ds"])
        return df_vacances

    school_holidays = pd.concat(map(_get_school_holidays_FR_for_year, years))
    holidays_df = school_holidays[["holiday", "ds"]]
    return holidays_df



from typing import Any, Dict

import streamlit as st


def input_metrics(readme: Dict[Any, Any], config: Dict[Any, Any]) -> Dict[Any, Any]:
    """Lets the user select evaluation metrics to be used for model evaluation.

    Parameters
    ----------
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    config : Dict
        Lib config containing the default list of metrics to use for evaluation.

    Returns
    -------
    dict
        Dictionary containing evaluation metrics information.
    """
    eval = dict()
    eval["metrics"] = st.multiselect(
        "Select evaluation metrics",
        ["MAPE", "SMAPE", "MSE", "RMSE", "MAE"],
        default=config["metrics"]["default"]["selection"],
        help=readme["tooltips"]["metrics"],
    )
    return eval


def input_scope_eval(eval: Dict[Any, Any], use_cv: bool, readme: Dict[Any, Any]) -> Dict[Any, Any]:
    """Lets the user define the scope of model evaluation (granularity, evaluation set).

    Parameters
    ----------
    eval : Dict
        Dictionary containing evaluation metrics information.
    use_cv : bool
        Whether or not cross-validation is used.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Dictionary containing information about the scope of model evaluation (granularity, evaluation set).
    """
    if use_cv:
        eval["set"] = "Validation"
        eval["granularity"] = "cutoff"
    else:
        eval["set"] = st.selectbox(
            "Select evaluation set", ["Validation", "Training"], help=readme["tooltips"]["eval_set"]
        )
        eval["granularity"] = st.selectbox(
            "Select evaluation granularity",
            ["Daily", "Day of Week", "Weekly", "Monthly", "Quarterly", "Yearly", "Global"],
            help=readme["tooltips"]["eval_granularity"],
        )
    eval["get_perf_on_agg_forecast"] = st.checkbox(
        "Get perf on aggregated forecast", value=False, help=readme["tooltips"]["choice_agg_perf"]
    )
    return eval



from typing import Any, Dict, Optional, Tuple

import pandas as pd
from fbprophet import Prophet
from fbprophet.diagnostics import cross_validation
from streamlit_prophet.lib.dataprep.clean import exp_transform
from streamlit_prophet.lib.dataprep.format import check_future_regressors_df
from streamlit_prophet.lib.dataprep.split import make_eval_df, make_future_df
from streamlit_prophet.lib.exposition.preparation import get_df_cv_with_hist
from streamlit_prophet.lib.models.preparation import add_prophet_holidays, get_prophet_cv_horizon
from streamlit_prophet.lib.utils.logging import suppress_stdout_stderr


def instantiate_prophet_model(
    params: Dict[Any, Any], use_regressors: bool = True, dates: Optional[Dict[Any, Any]] = None
) -> Prophet:
    """Instantiates a Prophet model with input parameters.

    Parameters
    ----------
    params : Dict
        Model parameters.
    use_regressors : bool
        Whether or not to add regressors to the model.
    dates : dict
        Dictionary containing all relevant dates for training and forecasting.

    Returns
    -------
    Prophet
        Instantiated (not fitted) Prophet model.
    """
    seasonality_params = {
        f"{k}_seasonality": params["seasonalities"][k]["prophet_param"]
        for k in {"yearly", "weekly", "daily"}.intersection(set(params["seasonalities"].keys()))
    }
    model = Prophet(**{**params["prior_scale"], **seasonality_params, **params["other"]})
    for _, values in params["seasonalities"].items():
        if "custom_param" in values:
            model.add_seasonality(**values["custom_param"])
    if dates:
        model = add_prophet_holidays(model, params["holidays"], dates)

    if use_regressors:
        for regressor in params["regressors"].keys():
            model.add_regressor(
                regressor, prior_scale=params["regressors"][regressor]["prior_scale"]
            )
    return model


def forecast_workflow(
    config: Dict[Any, Any],
    use_cv: bool,
    make_future_forecast: bool,
    evaluate: bool,
    cleaning: Dict[Any, Any],
    resampling: Dict[Any, Any],
    params: Dict[Any, Any],
    dates: Dict[Any, Any],
    datasets: Dict[Any, Any],
    df: pd.DataFrame,
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
    load_options: Dict[Any, Any],
) -> Tuple[Dict[Any, Any], Dict[Any, Any], Dict[Any, Any]]:
    """Trains a Prophet model and makes a prediction on evaluation data and future data if needed.

    Parameters
    ----------
    config : Dict
        Lib configuration dictionary, containing information about random seed to use for training.
    use_cv : bool
        Whether or not cross-validation is used.
    make_future_forecast : bool
        Whether or not to make a forecast on future dates.
    evaluate : bool
        Whether or not to do a model evaluation.
    cleaning : Dict
        Dataset cleaning specifications.
    resampling : Dict
        Dataset resampling specifications.
    params : Dict
        Model parameters.
    dates : Dict
        Dictionary containing all relevant dates for training and forecasting.
    datasets : Dict
        Dictionary containing all relevant dataframes for training and forecasting.
    df : pd.DataFrame
        Full input dataframe, after cleaning, filtering and resampling.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.
    load_options : Dict
        Loading options selected by user.

    Returns
    -------
    dict
        Dictionary containing all relevant dataframes for training and forecasting.
    dict
        Dictionary containing fitted Prophet models.
    dict
        Dictionary containing the different forecasts.
    """
    models: Dict[Any, Any] = dict()
    forecasts: Dict[Any, Any] = dict()
    with suppress_stdout_stderr():
        if evaluate:
            datasets, models, forecasts = forecast_eval(
                config, use_cv, resampling, params, dates, datasets, models, forecasts
            )
        if make_future_forecast:
            datasets, models, forecasts = forecast_future(
                config,
                params,
                cleaning,
                dates,
                datasets,
                models,
                forecasts,
                df,
                resampling,
                date_col,
                target_col,
                dimensions,
                load_options,
            )
    if cleaning["log_transform"] & (evaluate | make_future_forecast):
        datasets, forecasts = exp_transform(datasets, forecasts)
    return datasets, models, forecasts


def forecast_eval(
    config: Dict[Any, Any],
    use_cv: bool,
    resampling: Dict[Any, Any],
    params: Dict[Any, Any],
    dates: Dict[Any, Any],
    datasets: Dict[Any, Any],
    models: Dict[Any, Any],
    forecasts: Dict[Any, Any],
) -> Tuple[Dict[Any, Any], Dict[Any, Any], Dict[Any, Any]]:
    """Trains a Prophet model on training data and makes a prediction on evaluation data.

    Parameters
    ----------
    config : Dict
        Lib configuration dictionary, containing information about random seed to use for training.
    use_cv : bool
        Whether or not cross-validation is used.
    resampling : Dict
        Dataset resampling specifications.
    params : Dict
        Model parameters.
    dates : Dict
        Dictionary containing all relevant dates for training and forecasting.
    datasets : Dict
        Dictionary containing all relevant dataframes for training and forecasting.
    models : Dict
        Dictionary containing instantiated Prophet models.
    forecasts : Dict
        Dictionary containing the different forecasts.

    Returns
    -------
    dict
        Dictionary containing all relevant datasets for training and forecasting.
    dict
        Dictionary containing fitted Prophet models.
    dict
        Dictionary containing the different forecasts.
    """
    models["eval"] = instantiate_prophet_model(params, dates=dates)
    models["eval"].fit(datasets["train"], seed=config["global"]["seed"])
    if use_cv:
        forecasts["cv"] = cross_validation(
            models["eval"],
            cutoffs=dates["cutoffs"],
            horizon=get_prophet_cv_horizon(dates, resampling),
            parallel="processes",
        )
        forecasts["cv_with_hist"] = get_df_cv_with_hist(forecasts, datasets, models)
    else:
        datasets = make_eval_df(datasets)
        forecasts["eval"] = models["eval"].predict(datasets["eval"])
    return datasets, models, forecasts


def forecast_future(
    config: Dict[Any, Any],
    params: Dict[Any, Any],
    cleaning: Dict[Any, Any],
    dates: Dict[Any, Any],
    datasets: Dict[Any, Any],
    models: Dict[Any, Any],
    forecasts: Dict[Any, Any],
    df: pd.DataFrame,
    resampling: Dict[Any, Any],
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
    load_options: Dict[Any, Any],
) -> Tuple[Dict[Any, Any], Dict[Any, Any], Dict[Any, Any]]:
    """Trains a Prophet model on the whole dataset and makes a prediction on future data.

    Parameters
    ----------
    config : Dict
        Lib configuration dictionary, containing information about random seed to use for training.
    params : Dict
        Model parameters.
    cleaning : Dict
        Dataset cleaning specifications.
    dates : Dict
        Dictionary containing all relevant dates for training and forecasting.
    datasets : Dict
        Dictionary containing all relevant dataframes for training and forecasting.
    models : Dict
        Dictionary containing instantiated Prophet models.
    forecasts : Dict
        Dictionary containing the different forecasts.
    df : pd.DataFrame
        Full input dataframe, after cleaning, filtering and resampling.
    resampling : Dict
        Dictionary containing dataset frequency information.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.
    load_options : Dict
        Loading options selected by user.

    Returns
    -------
    dict
        Dictionary containing all relevant datasets for training and forecasting.
    dict
        Dictionary containing fitted Prophet models.
    dict
        Dictionary containing the different forecasts.
    """
    use_regressors = check_future_regressors_df(
        datasets, dates, params, resampling, date_col, dimensions
    )
    datasets = make_future_df(
        dates,
        df,
        datasets,
        cleaning,
        date_col,
        target_col,
        dimensions,
        load_options,
        config,
        resampling,
    )
    models["future"] = instantiate_prophet_model(params, use_regressors=use_regressors, dates=dates)
    models["future"].fit(datasets["full"], seed=config["global"]["seed"])
    forecasts["future"] = models["future"].predict(datasets["future"])
    return datasets, models, forecasts



# type: ignore[attr-defined]

import typer
from rich.console import Console
from streamlit_prophet import __version__
from streamlit_prophet.cli import deploy

app = typer.Typer(
    name="streamlit_prophet",
    help="`streamlit_prophet` is a Python cli/package",
    add_completion=True,
)
app.add_typer(deploy.app, name="deploy")
console = Console()


def version_callback(value: bool):
    """Prints the version of the package."""
    if value:
        console.print(f"[yellow]streamlit_prophet[/] version: [bold blue]{__version__}[/]")
        raise typer.Exit()



from typing import Any, Dict, List

import streamlit as st
from streamlit_prophet.lib.dataprep.clean import clean_df
from streamlit_prophet.lib.dataprep.format import (
    add_cap_and_floor_cols,
    check_dataset_size,
    filter_and_aggregate_df,
    format_date_and_target,
    format_datetime,
    print_empty_cols,
    print_removed_cols,
    remove_empty_cols,
    resample_df,
)
from streamlit_prophet.lib.dataprep.split import get_train_set, get_train_val_sets
from streamlit_prophet.lib.exposition.export import display_save_experiment_button
from streamlit_prophet.lib.exposition.visualize import (
    plot_components,
    plot_future,
    plot_overview,
    plot_performance,
)
from streamlit_prophet.lib.inputs.dataprep import input_cleaning, input_dimensions, input_resampling
from streamlit_prophet.lib.inputs.dataset import (
    input_columns,
    input_dataset,
    input_future_regressors,
)
from streamlit_prophet.lib.inputs.dates import (
    input_cv,
    input_forecast_dates,
    input_train_dates,
    input_val_dates,
)
from streamlit_prophet.lib.inputs.eval import input_metrics, input_scope_eval
from streamlit_prophet.lib.inputs.params import (
    input_holidays_params,
    input_other_params,
    input_prior_scale_params,
    input_regressors,
    input_seasonality_params,
)
from streamlit_prophet.lib.models.prophet import forecast_workflow
from streamlit_prophet.lib.utils.load import load_config, load_image

# Page config
st.set_page_config(page_title="Prophet", layout="wide")

# Load config
config, instructions, readme = load_config(
    "config_streamlit.toml", "config_instructions.toml", "config_readme.toml"
)

# Initialization
dates: Dict[Any, Any] = dict()
report: List[Dict[str, Any]] = []

# Info
with st.beta_expander("What is this app?", expanded=False):
    st.write(readme["app"]["app_intro"])
    st.write("")
st.write("")
st.sidebar.image(load_image("logo.png"), use_column_width=True)

st.sidebar.title("1. Data")

# Load data
with st.sidebar.beta_expander("Dataset", expanded=True):
    df, load_options, config, datasets = input_dataset(config, readme, instructions)
    df, empty_cols = remove_empty_cols(df)
    print_empty_cols(empty_cols)

# Column names
with st.sidebar.beta_expander("Columns", expanded=True):
    date_col, target_col = input_columns(config, readme, df, load_options)
    df = format_date_and_target(df, date_col, target_col, config, load_options)

# Filtering
with st.sidebar.beta_expander("Filtering", expanded=False):
    dimensions = input_dimensions(df, readme, config)
    df, cols_to_drop = filter_and_aggregate_df(df, dimensions, config, date_col, target_col)
    print_removed_cols(cols_to_drop)

# Resampling
with st.sidebar.beta_expander("Resampling", expanded=False):
    resampling = input_resampling(df, readme)
    df = format_datetime(df, resampling)
    df = resample_df(df, resampling)
    check_dataset_size(df, config)

# Cleaning
with st.sidebar.beta_expander("Cleaning", expanded=False):
    cleaning = input_cleaning(resampling, readme, config)
    df = clean_df(df, cleaning)
    check_dataset_size(df, config)

st.sidebar.title("2. Modelling")

# Prior scale
with st.sidebar.beta_expander("Prior scale", expanded=False):
    params = input_prior_scale_params(config, readme)

# Seasonalities
with st.sidebar.beta_expander("Seasonalities", expanded=False):
    params = input_seasonality_params(config, params, resampling, readme)

# Holidays
with st.sidebar.beta_expander("Holidays"):
    params = input_holidays_params(params, readme, config)

# External regressors
with st.sidebar.beta_expander("Regressors"):
    params = input_regressors(df, config, params, readme)

# Other parameters
with st.sidebar.beta_expander("Other parameters", expanded=False):
    params = input_other_params(config, params, readme)
    df = add_cap_and_floor_cols(df, params)

st.sidebar.title("3. Evaluation")

# Choose whether or not to do evaluation
evaluate = st.sidebar.checkbox(
    "Evaluate my model", value=True, help=readme["tooltips"]["choice_eval"]
)

if evaluate:

    # Split
    with st.sidebar.beta_expander("Split", expanded=True):
        use_cv = st.checkbox(
            "Perform cross-validation", value=False, help=readme["tooltips"]["choice_cv"]
        )
        dates = input_train_dates(df, use_cv, config, resampling, dates)
        if use_cv:
            dates = input_cv(dates, resampling, config, readme)
            datasets = get_train_set(df, dates, datasets)
        else:
            dates = input_val_dates(df, dates, config)
            datasets = get_train_val_sets(df, dates, config, datasets)

    # Performance metrics
    with st.sidebar.beta_expander("Metrics", expanded=False):
        eval = input_metrics(readme, config)

    # Scope of evaluation
    with st.sidebar.beta_expander("Scope", expanded=False):
        eval = input_scope_eval(eval, use_cv, readme)

else:
    use_cv = False

st.sidebar.title("4. Forecast")

# Choose whether or not to do future forecasts
make_future_forecast = st.sidebar.checkbox(
    "Make forecast on future dates", value=False, help=readme["tooltips"]["choice_forecast"]
)
if make_future_forecast:
    with st.sidebar.beta_expander("Horizon", expanded=False):
        dates = input_forecast_dates(df, dates, resampling, config, readme)
    with st.sidebar.beta_expander("Regressors", expanded=False):
        datasets = input_future_regressors(
            datasets, dates, params, dimensions, load_options, date_col
        )

# Launch training & forecast
if st.checkbox(
    "Launch forecast",
    value=False,
    help=readme["tooltips"]["launch_forecast"],
):

    if not (evaluate | make_future_forecast):
        st.error("Please check at least 'Evaluation' or 'Forecast' in the sidebar.")

    track_experiments = st.checkbox(
        "Track experiments", value=False, help=readme["tooltips"]["track_experiments"]
    )

    datasets, models, forecasts = forecast_workflow(
        config,
        use_cv,
        make_future_forecast,
        evaluate,
        cleaning,
        resampling,
        params,
        dates,
        datasets,
        df,
        date_col,
        target_col,
        dimensions,
        load_options,
    )

    # Visualizations

    if evaluate | make_future_forecast:
        st.write("# 1. Overview")
        report = plot_overview(
            make_future_forecast, use_cv, models, forecasts, target_col, cleaning, readme, report
        )

    if evaluate:
        st.write(
            f'# 2. Evaluation on {"CV" if use_cv else ""} {eval["set"].lower()} set{"s" if use_cv else ""}'
        )
        report = plot_performance(
            use_cv, target_col, datasets, forecasts, dates, eval, resampling, config, readme, report
        )

    if evaluate | make_future_forecast:
        st.write(
            "# 3. Impact of components and regressors"
            if evaluate
            else "# 2. Impact of components and regressors"
        )
        report = plot_components(
            use_cv,
            make_future_forecast,
            target_col,
            models,
            forecasts,
            cleaning,
            resampling,
            config,
            readme,
            df,
            report,
        )

    if make_future_forecast:
        st.write("# 4. Future forecast" if evaluate else "# 3. Future forecast")
        report = plot_future(models, forecasts, dates, target_col, cleaning, readme, report)

    # Save experiment
    if track_experiments:
        display_save_experiment_button(
            report,
            config,
            use_cv,
            make_future_forecast,
            evaluate,
            cleaning,
            resampling,
            params,
            dates,
            date_col,
            target_col,
            dimensions,
        )



from typing import Any, Dict

import pandas as pd
from fbprophet import Prophet
from streamlit_prophet.lib.utils.holidays import lockdown_format_func
from streamlit_prophet.lib.utils.mapping import (
    COVID_LOCKDOWN_DATES_MAPPING,
    SCHOOL_HOLIDAYS_FUNC_MAPPING,
    convert_into_nb_of_days,
    convert_into_nb_of_seconds,
)


def get_prophet_cv_horizon(dates: Dict[Any, Any], resampling: Dict[Any, Any]) -> str:
    """Returns cross-validation horizon at the right format for Prophet cross_validation function.

    Parameters
    ----------
    dates : Dict
        Dictionary containing cross-validation information.
    resampling : Dict
        Dictionary containing dataset frequency information.

    Returns
    -------
    str
        Cross-validation horizon at the right format for Prophet cross_validation function.
    """
    freq = resampling["freq"][-1]
    horizon = dates["folds_horizon"]
    if freq in ["s", "H"]:
        prophet_horizon = f"{convert_into_nb_of_seconds(freq, horizon)} seconds"
    else:
        prophet_horizon = f"{convert_into_nb_of_days(freq, horizon)} days"
    return prophet_horizon


def add_prophet_holidays(
    model: Prophet, holidays_params: Dict[Any, Any], dates: Dict[Any, Any]
) -> pd.DataFrame:
    """Add all available holidays to the Prophet model

    Parameters
    ----------
    model: Prophet
        Prophet model to add holidays to
    holidays_params: dict
        dict of parameters including 'country': str, 'public_holidays': bool, 'school_holidays': bool, lockdown_events: List[int]
    dates : dict
        Dictionary containing all relevant dates for training and forecasting.

    Returns
    -------
    Prophet
        Prophet model with holidays added
    """
    country = holidays_params["country"]
    if holidays_params["public_holidays"]:
        model.add_country_holidays(country)

    holidays_df_list = []
    if holidays_params["school_holidays"]:
        all_dates = {
            k: v
            for k, v in dates.items()
            if k not in ["n_folds", "folds_horizon", "forecast_horizon", "cutoffs", "forecast_freq"]
        }
        years = list(range(min(all_dates.values()).year, max(all_dates.values()).year + 1))
        get_holidays_func = SCHOOL_HOLIDAYS_FUNC_MAPPING[country]
        holidays_df = get_holidays_func(years)
        holidays_df[["lower_window", "upper_window"]] = 0
        holidays_df_list.append(holidays_df)

    for lockdown_idx in holidays_params["lockdown_events"]:
        start, end = COVID_LOCKDOWN_DATES_MAPPING[country][lockdown_idx]
        lockdown_df = pd.DataFrame(
            {
                "holiday": lockdown_format_func(lockdown_idx),
                "ds": pd.date_range(start=start, end=end),
                "lower_window": 0,
                "upper_window": 0,
            }
        )
        holidays_df_list.append(lockdown_df)

    if len(holidays_df_list) == 0:
        return model
    holidays_df = pd.concat(holidays_df_list, sort=True)
    model.holidays = holidays_df
    return model



from typing import Any, Dict, Tuple

import io
from pathlib import Path

import pandas as pd
import requests
import streamlit as st
import toml
from PIL import Image


def get_project_root() -> str:
    """Returns project root path.

    Returns
    -------
    str
        Project root path.
    """
    return str(Path(__file__).parent.parent.parent)


@st.cache(suppress_st_warning=True)
def load_dataset(file: str, load_options: Dict[Any, Any]) -> pd.DataFrame:
    """Loads dataset from user's file system as a pandas dataframe.

    Parameters
    ----------
    file
        Uploaded dataset file.
    load_options : Dict
        Dictionary containing separator information.

    Returns
    -------
    pd.DataFrame
        Loaded dataset.
    """
    try:
        return pd.read_csv(file, sep=load_options["separator"])
    except:
        st.error(
            "This file can't be converted into a dataframe. Please import a csv file with a valid separator."
        )
        st.stop()


@st.cache(allow_output_mutation=True)
def load_config(
    config_streamlit_filename: str, config_instructions_filename: str, config_readme_filename: str
) -> Tuple[Dict[Any, Any], Dict[Any, Any], Dict[Any, Any]]:
    """Loads configuration files.

    Parameters
    ----------
    config_streamlit_filename : str
        Filename of lib configuration file.
    config_instructions_filename : str
        Filename of custom config instruction file.
    config_readme_filename : str
        Filename of readme configuration file.

    Returns
    -------
    dict
        Lib configuration file.
    dict
        Readme configuration file.
    """
    config_streamlit = toml.load(Path(get_project_root()) / f"config/{config_streamlit_filename}")
    config_instructions = toml.load(
        Path(get_project_root()) / f"config/{config_instructions_filename}"
    )
    config_readme = toml.load(Path(get_project_root()) / f"config/{config_readme_filename}")
    return dict(config_streamlit), dict(config_instructions), dict(config_readme)


@st.cache()
def download_toy_dataset(url: str) -> pd.DataFrame:
    """Downloads a toy dataset from an external source and converts it into a pandas dataframe.

    Parameters
    ----------
    url : str
        Link to the toy dataset.

    Returns
    -------
    pd.DataFrame
        Loaded dataset.
    """
    download = requests.get(url).content
    df = pd.read_csv(io.StringIO(download.decode("utf-8")))
    return df


@st.cache()
def load_custom_config(config_file: io.BytesIO) -> Dict[Any, Any]:
    """Loads config toml file from user's file system as a dictionary.

    Parameters
    ----------
    config_file
        Uploaded toml config file.

    Returns
    -------
    dict
        Loaded config dictionary.
    """
    toml_file = Path(get_project_root()) / f"config/custom_{config_file.name}"
    write_bytesio_to_file(str(toml_file), config_file)
    config = toml.load(toml_file)
    return dict(config)


def write_bytesio_to_file(filename: str, bytesio: io.BytesIO) -> None:
    """
    Write the contents of the given BytesIO to a file.

    Parameters
    ----------
    filename
        Uploaded toml config file.
    bytesio
        BytesIO object.
    """
    with open(filename, "wb") as outfile:
        outfile.write(bytesio.getbuffer())


@st.cache
def load_image(image_name: str) -> Image:
    """Displays an image.

    Parameters
    ----------
    image_name : str
        Local path of the image.

    Returns
    -------
    Image
        Image to be displayed.
    """
    return Image.open(Path(get_project_root()) / f"references/{image_name}")



from typing import Any, Dict, List

import pandas as pd
import streamlit as st
from streamlit_prophet.lib.utils.mapping import dayname_to_daynumber


def input_cleaning(
    resampling: Dict[Any, Any], readme: Dict[Any, Any], config: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user enter cleaning specifications.

    Parameters
    ----------
    resampling : Dict
        Dictionary containing dataset frequency information.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    config : Dict
        Dictionary where user can provide default cleaning choices.

    Returns
    -------
    dict
        Cleaning specifications (remove_days, del_days, del_negative, del_zeros, log_transform).
    """
    cleaning: Dict[Any, Any] = dict()
    if resampling["freq"][-1] in ["s", "H", "D"]:
        del_days = st.multiselect(
            "Remove days",
            ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"],
            default=config["dataprep"]["remove_days"],
            help=readme["tooltips"]["remove_days"],
        )
        cleaning["del_days"] = dayname_to_daynumber(del_days)
    else:
        cleaning["del_days"] = []
    cleaning["del_zeros"] = st.checkbox(
        "Delete rows where target = 0",
        False if config["dataprep"]["remove_zeros"] in ["false", False] else True,
        help=readme["tooltips"]["del_zeros"],
    )
    cleaning["del_negative"] = st.checkbox(
        "Delete rows where target < 0",
        False if config["dataprep"]["remove_negative"] in ["false", False] else True,
        help=readme["tooltips"]["del_negative"],
    )
    cleaning["log_transform"] = st.checkbox(
        "Target log transform",
        False if config["dataprep"]["log_transform"] in ["false", False] else True,
        help=readme["tooltips"]["log_transform"],
    )
    return cleaning


def input_dimensions(
    df: pd.DataFrame, readme: Dict[Any, Any], config: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user enter filtering and aggregation specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be used to detect dimension columns.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    config : Dict
        Dictionary where user can provide the list of dimensions.

    Returns
    -------
    dict
        Filtering and aggregation specifications (dimensions, values to keep, aggregation function).
    """
    dimensions: Dict[Any, Any] = dict()
    eligible_cols = sorted(set(df.columns) - {"ds", "y"})
    if len(eligible_cols) > 0:
        config_dimensions = config["columns"]["dimensions"]
        if config_dimensions not in ["false", False]:
            if len(set(config_dimensions).intersection(set(eligible_cols))) != len(
                config_dimensions
            ):
                st.error(
                    f"Selected dimensions are not in the dataset columns, "
                    f"please provide a list of valid columns for dimensions in the config file."
                )
                st.stop()
        dimensions_cols = st.multiselect(
            "Select dataset dimensions if any",
            list(eligible_cols),
            default=_autodetect_dimensions(df)
            if config_dimensions in ["false", False]
            else config_dimensions,
            help=readme["tooltips"]["dimensions"],
        )
        for col in dimensions_cols:
            values = list(df[col].unique())
            if st.checkbox(
                f"Keep all values for {col}",
                True,
                help=readme["tooltips"]["dimensions_keep"] + col + ".",
            ):
                dimensions[col] = values.copy()
            else:
                dimensions[col] = st.multiselect(
                    f"Values to keep for {col}",
                    values,
                    default=[values[0]],
                    help=readme["tooltips"]["dimensions_filter"],
                )
        dimensions["agg"] = st.selectbox(
            "Target aggregation function over dimensions",
            config["dataprep"]["dimensions_agg"],
            help=readme["tooltips"]["dimensions_agg"],
        )
    else:
        st.write("Date and target are the only columns in your dataset, there are no dimensions.")
        dimensions["agg"] = "Mean"
    return dimensions


def _autodetect_dimensions(df: pd.DataFrame) -> List[Any]:
    """Detects dimension columns in input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be used to detect dimension columns.

    Returns
    -------
    list
        List of dimension columns detected. The user will be able to change that list later if it is incorrect.
    """
    eligible_cols = sorted(set(df.columns) - {"ds", "y"})
    detected_cols = []
    for col in eligible_cols:
        values = df[col].value_counts()
        values = values.loc[values > 0].to_list()
        if (len(values) > 1) & (len(values) < 0.05 * len(df)):
            if max(values) / min(values) <= 20:
                detected_cols.append(col)
    return detected_cols


def input_resampling(df: pd.DataFrame, readme: Dict[Any, Any]) -> Dict[Any, Any]:
    """Lets the user enter resampling specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be used to detect current frequency in dataset.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Resampling specifications (resample or not, frequency, aggregation function).
    """
    resampling: Dict[Any, Any] = dict()
    resampling["freq"] = _autodetect_freq(df)
    st.write(f"Frequency detected in dataset: {resampling['freq']}")
    resampling["resample"] = st.checkbox(
        "Resample my dataset", False, help=readme["tooltips"]["resample_choice"]
    )
    if resampling["resample"]:
        current_freq = resampling["freq"][-1]
        possible_freq_names = ["Hourly", "Daily", "Weekly", "Monthly", "Quarterly", "Yearly"]
        possible_freq = [freq[0] for freq in possible_freq_names]
        current_freq_index = possible_freq.index(current_freq)
        if current_freq != "Y":
            new_freq = st.selectbox(
                "Select new frequency",
                possible_freq_names[current_freq_index + 1 :],
                help=readme["tooltips"]["resample_new_freq"],
            )
            resampling["freq"] = new_freq[0]
            resampling["agg"] = st.selectbox(
                "Target aggregation function when resampling",
                ["Mean", "Sum", "Max", "Min"],
                help=readme["tooltips"]["resample_agg"],
            )
        else:
            st.write("Frequency is already yearly, resampling is not possible.")
            resampling["resample"] = False
    return resampling


def _autodetect_freq(df: pd.DataFrame) -> str:
    """Detects date frequency of input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be used to detect dataset frequency.

    Returns
    -------
    str
        Frequency detected. The user will be able to resample later if it is not the value expected.
    """
    min_delta = pd.Series(df["ds"]).diff().min()
    days = min_delta.days
    seconds = min_delta.seconds
    if days == 1:
        return "D"
    elif days < 1:
        if seconds >= 3600:
            return f"{round(seconds/3600)}H"
        else:
            return f"{seconds}s"
    elif days > 1:
        if days < 7:
            return f"{days}D"
        elif days < 28:
            return f"{round(days/7)}W"
        elif days < 90:
            return f"{round(days/30)}M"
        elif days < 365:
            return f"{round(days/90)}Q"
        else:
            return f"{round(days/365)}Y"
    else:
        raise ValueError("No frequency detected.")



from typing import Any, Dict

import plotly.graph_objects as go
import streamlit as st
from streamlit_prophet.lib.exposition.preparation import get_cv_dates_dict, get_hover_template_cv


def plot_cv_dates(
    cv_dates: Dict[Any, Any], resampling: Dict[Any, Any], style: Dict[Any, Any]
) -> go.Figure:
    """Creates a plotly bar plot showing training and validation dates for each cross-validation fold.

    Parameters
    ----------
    cv_dates : Dict
        Dictionary containing training and validation dates of each cross-validation fold.
    resampling : Dict
        Resampling specifications (granularity, dataset frequency).
    style : Dict
        Style specifications for the graph (colors).

    Returns
    -------
    go.Figure
        Plotly bar plot showing training and validation dates for each cross-validation fold.
    """
    hover_data, hover_template = get_hover_template_cv(cv_dates, resampling)
    fig = go.Figure()
    fig.add_trace(
        go.Bar(
            y=list(cv_dates.keys()),
            x=[cv_dates[fold]["val_end"] for fold in cv_dates.keys()],
            name="",
            orientation="h",
            text=hover_data,
            hoverinfo="y+text",
            hovertemplate=hover_template,
            marker=dict(color=style["colors"][1], line=dict(color=style["colors"][1], width=2)),
        )
    )
    fig.add_trace(
        go.Bar(
            y=list(cv_dates.keys()),
            x=[cv_dates[fold]["train_start"] for fold in cv_dates.keys()],
            name="",
            orientation="h",
            text=hover_data,
            hoverinfo="y+text",
            hovertemplate=hover_template,
            marker=dict(color=style["colors"][0], line=dict(color=style["colors"][1], width=2)),
        )
    )
    fig.add_trace(
        go.Bar(
            y=list(cv_dates.keys()),
            x=[cv_dates[fold]["train_end"] for fold in cv_dates.keys()],
            name="",
            orientation="h",
            text=hover_data,
            hoverinfo="y+text",
            hovertemplate=hover_template,
            marker=dict(color=style["colors"][0], line=dict(color=style["colors"][1], width=2)),
        )
    )
    fig.update_layout(
        showlegend=False,
        barmode="overlay",
        xaxis_type="date",
        title_text="Cross-Validation Folds",
        title_x=0.5,
        title_y=0.85,
    )
    return fig


def display_expander(
    readme: Dict[Any, Any], section: str, title: str, add_blank: bool = False
) -> None:
    """Displays a streamlit expander with information about a section of the dashboard.

    Parameters
    ----------
    readme : Dict
        Dictionary containing explanations about the section.
    section : str
        Section of the dashboard on top of which the expander will be displayed.
    title : str
        Title for the expander.
    add_blank : bool
        Whether or not to add a blank after the expander.
    """
    with st.beta_expander(title, expanded=False):
        st.write(readme["plots"][section])
        st.write("")
    if add_blank:
        st.write("")
        st.write("")


def display_expanders_performance(
    use_cv: bool,
    dates: Dict[Any, Any],
    resampling: Dict[Any, Any],
    style: Dict[Any, Any],
    readme: Dict[Any, Any],
) -> None:
    """Displays a streamlit expander with information about performance section.

    Parameters
    ----------
    use_cv : bool
        Whether or not cross-validation is used.
    dates : Dict
        Dictionary containing cross-validation dates information.
    resampling : Dict
        Resampling specifications (granularity, dataset frequency).
    style : Dict
        Style specifications for the graph (colors).
    readme : Dict
        Dictionary containing explanations about the section.
    """
    st.write("")
    with st.beta_expander("More info on evaluation metrics", expanded=False):
        st.write(readme["plots"]["metrics"])
        st.write("")
        _display_metrics()
        st.write("")
    if use_cv:
        cv_dates = get_cv_dates_dict(dates, resampling)
        with st.beta_expander("See cross-validation folds", expanded=False):
            st.plotly_chart(plot_cv_dates(cv_dates, resampling, style))


def _display_metrics() -> None:
    """Displays formulas for all performance metrics."""
    if st.checkbox("Show metric formulas", value=False):
        st.write("If N is the number of distinct dates in the evaluation set:")
        st.latex(r"MAPE = \dfrac{1}{N}\sum_{t=1}^{N}|\dfrac{Truth_t - Forecast_t}{Truth_t}|")
        st.latex(r"RMSE = \sqrt{\dfrac{1}{N}\sum_{t=1}^{N}(Truth_t - Forecast_t)^2}")
        st.latex(
            r"SMAPE = \dfrac{1}{N}\sum_{t=1}^{N}\dfrac{2|Truth_t - Forecast_t]}{|Truth_t| + |Forecast_t|}"
        )
        st.latex(r"MSE = \dfrac{1}{N}\sum_{t=1}^{N}(Truth_t - Forecast_t)^2")
        st.latex(r"MAE = \dfrac{1}{N}\sum_{t=1}^{N}|Truth_t - Forecast_t|")



from typing import Any, Dict, List

import base64
import io
import re
import uuid
from base64 import b64encode
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import pandas as pd
import plotly.graph_objects as go
import streamlit as st
import toml
from streamlit_prophet.lib.utils.load import get_project_root


def get_dataframe_download_link(df: pd.DataFrame, filename: str, linkname: str) -> str:
    """Creates a link to download a dataframe as a csv file.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to export.
    filename : str
        Name of the exported file.
    linkname : str
        Text displayed in the streamlit app.

    Returns
    -------
    str
        Download link.
    """
    csv = df.to_csv()
    b64 = base64.b64encode(csv.encode()).decode()
    href = f'<a href="data:file/csv;base64,{b64}" download="{filename}.csv">{linkname}</a>'
    return href


def get_config_download_link(config: Dict[Any, Any], filename: str, linkname: str) -> str:
    """Creates a link to download the config as a toml file. Removes keys that should not be customized.

    Parameters
    ----------
    config : Dict
        Config file to export as toml.
    filename : str
        Name of the exported file.
    linkname : str
        Text displayed in the streamlit app.

    Returns
    -------
    str
        Download link.
    """
    config_template = config.copy()
    if "datasets" in config_template.keys():
        del config_template["datasets"]
    toml_string = toml.dumps(config)
    b64 = base64.b64encode(toml_string.encode()).decode()
    href = f'<a href="data:file/toml;base64,{b64}" download="{filename}">{linkname}</a>'
    return href


def get_plotly_download_link(fig: go.Figure, filename: str, linkname: str) -> str:
    """Creates a link to download a dataframe as a html file.

    Parameters
    ----------
    fig : go.Figure
        Plotly go figure to export.
    filename : str
        Name of the exported file.
    linkname : str
        Text displayed in the streamlit app.

    Returns
    -------
    str
        Download link.
    """
    buffer = io.StringIO()
    fig.write_html(buffer)
    html_bytes = buffer.getvalue().encode()
    encoded = b64encode(html_bytes).decode()
    href = f'<a href="data:text/html;base64,{encoded}" download="{filename}.html">{linkname}</a>'
    return href


def display_dataframe_download_link(
    df: pd.DataFrame, filename: str, linkname: str, add_blank: bool = False
) -> None:
    """Displays a link to download a dataframe as a csv file.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to export.
    filename : str
        Name of the exported file.
    linkname : str
        Text displayed in the streamlit app.
    add_blank : str
        Whether or not to add a blank before the link in streamlit app.
    """
    if add_blank:
        st.write("")
    st.markdown(get_dataframe_download_link(df, filename, linkname), unsafe_allow_html=True)


def display_2_dataframe_download_links(
    df1: pd.DataFrame,
    filename1: str,
    linkname1: str,
    df2: pd.DataFrame,
    filename2: str,
    linkname2: str,
    add_blank: bool = False,
) -> None:
    """Displays a link to download a dataframe as a csv file.

    Parameters
    ----------
    df1 : pd.DataFrame
        First dataframe to export.
    filename1 : str
        Name of the first exported file.
    linkname1 : str
        Text displayed in the streamlit app for the first link.
    df2 : pd.DataFrame
        Second dataframe to export.
    filename2 : str
        Name of the second exported file.
    linkname2 : str
        Text displayed in the streamlit app for the second link.
    add_blank : str
        Whether or not to add a blank before the link in streamlit app.
    """
    if add_blank:
        st.write("")
    col1, col2 = st.beta_columns(2)
    col1.markdown(
        f"<p style='text-align: center;;'> {get_dataframe_download_link(df1, filename1, linkname1)}</p>",
        unsafe_allow_html=True,
    )
    col2.markdown(
        f"<p style='text-align: center;;'> {get_dataframe_download_link(df2, filename2, linkname2)}</p>",
        unsafe_allow_html=True,
    )


def display_config_download_links(
    config1: Dict[Any, Any],
    filename1: str,
    linkname1: str,
    config2: Dict[Any, Any],
    filename2: str,
    linkname2: str,
) -> None:
    """Displays a link to download a dataframe as a csv file.

    Parameters
    ----------
    config1 : Dict
        First config file to export as toml.
    filename1 : str
        Name of the first exported file.
    linkname1 : str
        Text displayed in the streamlit app for the first link.
    config2 : Dict
        Second config file to export as toml.
    filename2 : str
        Name of the second exported file.
    linkname2 : str
        Text displayed in the streamlit app for the second link.
    """
    col1, col2 = st.beta_columns(2)
    col1.markdown(
        f"<p style='text-align: center;;'> {get_config_download_link(config1, filename1, linkname1)}</p>",
        unsafe_allow_html=True,
    )
    col2.markdown(
        f"<p style='text-align: center;;'> {get_config_download_link(config2, filename2, linkname2)}</p>",
        unsafe_allow_html=True,
    )


def display_plotly_download_link(
    fig: go.Figure, filename: str, linkname: str, add_blank: bool = False
) -> None:
    """Displays a link to export a plotly graph in html.

    Parameters
    ----------
    fig : go.Figure
        Plotly go figure to export.
    filename : str
        Name of the exported file.
    linkname : str
        Text displayed in the streamlit app.
    add_blank : str
        Whether or not to add a blank before the link in streamlit app.
    """
    if add_blank:
        st.write("")
    st.markdown(get_plotly_download_link(fig, filename, linkname), unsafe_allow_html=True)


def create_report_zip_file(
    report: List[Dict[str, Any]],
    config: Dict[Any, Any],
    use_cv: bool,
    make_future_forecast: bool,
    evaluate: bool,
    cleaning: Dict[Any, Any],
    resampling: Dict[Any, Any],
    params: Dict[Any, Any],
    dates: Dict[Any, Any],
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
) -> str:
    """Saves locally all report components in a zip file.

    Parameters
    ----------
    report: List[Dict[str, Any]]
        List of all report components.
    config : Dict
        Lib configuration dictionary.
    use_cv : bool
        Whether or not cross-validation is used.
    make_future_forecast : bool
        Whether or not to make a forecast on future dates.
    evaluate : bool
        Whether or not to do a model evaluation.
    cleaning : Dict
        Dataset cleaning specifications.
    resampling : Dict
        Dataset resampling specifications.
    params : Dict
        Model parameters.
    dates : Dict
        Dictionary containing all relevant dates for training and forecasting.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.

    Returns
    -------
    str
        Path of the zip file.
    """
    # Create zip file
    zip_path = "experiment.zip"
    zipObj = ZipFile(zip_path, "w")
    report_name = f"report_{datetime.now().strftime('%Y%m%d_%Hh%Mm%Ss')}"
    # Save plots and data
    for x in report:
        if x["type"] == "plot":
            file_name = f"{report_name}/plots/{x['name']}.html"
            file_path = _get_file_path(file_name)
            x["object"].write_html(file_path)
        if x["type"] == "dataset":
            file_name = f"{report_name}/data/{x['name']}.csv"
            file_path = _get_file_path(file_name)
            x["object"].to_csv(file_path, index=False)
        zipObj.write(file_path, arcname=file_name)
    # Save default config
    default_config = config.copy()
    if "datasets" in default_config.keys():
        del default_config["datasets"]
    file_name = f"{report_name}/config/default_config.toml"
    file_path = _get_file_path(file_name)
    with open(file_path, "w") as toml_file:
        toml.dump(default_config, toml_file)
    zipObj.write(file_path, arcname=file_name)
    # Save user specifications
    all_specs = {
        "model_params": params,
        "dates": dates,
        "columns": {"date": date_col, "target": target_col},
        "filtering": dimensions,
        "cleaning": cleaning,
        "resampling": resampling,
        "actions": {
            "evaluate": evaluate,
            "use_cv": use_cv,
            "make_future_forecast": make_future_forecast,
        },
    }
    file_name = f"{report_name}/config/user_specifications.toml"
    file_path = _get_file_path(file_name)
    with open(file_path, "w") as toml_file:
        toml.dump(all_specs, toml_file)
    zipObj.write(file_path, arcname=file_name)
    # Close zip file
    zipObj.close()
    return zip_path


def _get_file_path(file_name: str) -> str:
    """Returns the full path of a file to include in the zip file.

    Parameters
    ----------
    file_name: str
        Short file name.

    Returns
    -------
    str
        Full path.
    """
    return str(Path(get_project_root()) / f"report/{'/'.join(file_name.split('/')[1:])}")


def create_save_experiment_button(zip_path: str) -> None:
    """Displays a link to export the report as a zip file.

    Parameters
    ----------
    zip_path: str
        Path of the zip file.
    """

    button_uuid = str(uuid.uuid4()).replace("-", "")
    button_id = re.sub(r"\d+", "", button_uuid)

    with open(zip_path, "rb") as f:
        bytes = f.read()
        b64 = base64.b64encode(bytes).decode()
        href = (
            f"<a download='{zip_path}' id='{button_id}' href=\"data:file/zip;base64,{b64}\" >Save "
            f"experiment</a><br></br> "
        )

    color1 = "rgb(255, 0, 102)"
    color2 = "rgb(0, 34, 68)"
    color3 = "rgb(255, 255, 255)"
    custom_css = f"""
            <style>
                #{button_id} {{
                    background-color: {color1};
                    color: {color3};
                    padding: 0.45em 0.58em;
                    position: relative;
                    text-decoration: none;
                    border-radius: 5px;
                    border-width: 2px;
                    border-style: solid;
                    border-color: {color3};
                    border-image: initial;
                }}
                #{button_id}:hover {{
                    border-color: {color2};
                    color: {color2};
                }}
                #{button_id}:active {{
                    box-shadow: none;
                    background-color: {color3};
                    border-color: {color1};
                    color: {color1};
                    }}
            </style> """

    st.markdown(
        f"<p style='text-align: center;;'> {custom_css + href} </p>", unsafe_allow_html=True
    )


def display_save_experiment_button(
    report: List[Dict[str, Any]],
    config: Dict[Any, Any],
    use_cv: bool,
    make_future_forecast: bool,
    evaluate: bool,
    cleaning: Dict[Any, Any],
    resampling: Dict[Any, Any],
    params: Dict[Any, Any],
    dates: Dict[Any, Any],
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
) -> None:
    """Saves locally all report components in a zip file.

    Parameters
    ----------
    report: List[Dict[str, Any]]
        List of all report components.
    config : Dict
        Lib configuration dictionary.
    use_cv : bool
        Whether or not cross-validation is used.
    make_future_forecast : bool
        Whether or not to make a forecast on future dates.
    evaluate : bool
        Whether or not to do a model evaluation.
    cleaning : Dict
        Dataset cleaning specifications.
    resampling : Dict
        Dataset resampling specifications.
    params : Dict
        Model parameters.
    dates : Dict
        Dictionary containing all relevant dates for training and forecasting.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.
    """
    with st.spinner("Saving config, plots and data..."):
        zip_path = create_report_zip_file(
            report,
            config,
            use_cv,
            make_future_forecast,
            evaluate,
            cleaning,
            resampling,
            params,
            dates,
            date_col,
            target_col,
            dimensions,
        )
        create_save_experiment_button(zip_path)



from typing import Any, Dict, Tuple

from datetime import timedelta

import numpy as np
import pandas as pd
from streamlit_prophet.lib.evaluation.preparation import add_time_groupers
from streamlit_prophet.lib.utils.mapping import convert_into_nb_of_days, convert_into_nb_of_seconds


def MAPE(y_true: pd.Series, y_pred: pd.Series) -> float:
    """Computes Mean Absolute Percentage Error (MAPE). Must be multiplied by 100 to get percentage.

    Parameters
    ----------
    y_true : pd.Series
        Ground truth target series.
    y_pred : pd.Series
        Prediction series.

    Returns
    -------
    float
        Mean Absolute Percentage Error (MAPE).
    """
    try:
        y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
        mask = (y_true != 0) & (~np.isnan(y_true)) & (~np.isnan(y_pred))
        y_true, y_pred = y_true, y_pred
        mape = np.mean(np.abs((y_true - y_pred) / y_true)[mask])
        return 0 if np.isnan(mape) else float(mape)
    except:
        return 0


def SMAPE(y_true: pd.Series, y_pred: pd.Series) -> float:
    """Computes Symmetric Mean Absolute Percentage Error (SMAPE).

    Parameters
    ----------
    y_true : pd.Series
        Ground truth target series.
    y_pred : pd.Series
        Prediction series.

    Returns
    -------
    float
        Symmetric Mean Absolute Percentage Error (SMAPE).
    """
    try:
        y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
        mask = (abs(y_true) + abs(y_pred) != 0) & (~np.isnan(y_true)) & (~np.isnan(y_pred))
        y_true, y_pred = y_true, y_pred
        nominator = np.abs(y_true - y_pred)
        denominator = np.abs(y_true) + np.abs(y_pred)
        smape = np.mean((2.0 * nominator / denominator)[mask])
        return 0 if np.isnan(smape) else float(smape)
    except:
        return 0


def MSE(y_true: pd.Series, y_pred: pd.Series) -> float:
    """Computes Mean Squared Error (MSE).

    Parameters
    ----------
    y_true : pd.Series
        Ground truth target series.
    y_pred : pd.Series
        Prediction series.

    Returns
    -------
    float
        Mean Squared Error (MSE).
    """
    try:
        y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
        mask = (~np.isnan(y_true)) & (~np.isnan(y_pred))
        mse = ((y_true - y_pred) ** 2)[mask].mean()
        return 0 if np.isnan(mse) else float(mse)
    except:
        return 0


def RMSE(y_true: pd.Series, y_pred: pd.Series) -> float:
    """Computes Root Mean Squared Error (RMSE).

    Parameters
    ----------
    y_true : pd.Series
        Ground truth target series.
    y_pred : pd.Series
        Prediction series.

    Returns
    -------
    float
        Root Mean Squared Error (RMSE).
    """
    y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
    rmse = np.sqrt(MSE(y_true, y_pred))
    return float(rmse)


def MAE(y_true: pd.Series, y_pred: pd.Series) -> float:
    """Computes Mean Absolute Error (MAE).

    Parameters
    ----------
    y_true : pd.Series
        Ground truth target series.
    y_pred : pd.Series
        Prediction series.

    Returns
    -------
    float
        Mean Absolute Error (MAE).
    """
    try:
        y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
        mask = (~np.isnan(y_true)) & (~np.isnan(y_pred))
        mae = abs(y_true - y_pred)[mask].mean()
        return 0 if np.isnan(mae) else float(mae)
    except:
        return 0


def get_perf_metrics(
    evaluation_df: pd.DataFrame,
    eval: Dict[Any, Any],
    dates: Dict[Any, Any],
    resampling: Dict[Any, Any],
    use_cv: bool,
    config: Dict[Any, Any],
) -> Tuple[pd.DataFrame, Dict[Any, Any]]:
    """Computes all metrics to gather them in a dataframe and a dictionary.

    Parameters
    ----------
    evaluation_df : pd.DataFrame
        Evaluation dataframe.
    eval : Dict
        Evaluation specifications.
    dates : Dict
        Dictionary containing all dates information.
    resampling : Dict
        Resampling specifications.
    use_cv : bool
        Whether or note cross-validation is used.
    config : Dict
        Lib configuration dictionary.

    Returns
    -------
    pd.DataFrame
        Dataframe with all metrics at the desired granularity.
    dict
        Dictionary with all metrics at the desired granularity.
    """
    df = _preprocess_eval_df(evaluation_df, use_cv)
    metrics_df = _compute_metrics(df, eval)
    metrics_df, metrics_dict = _format_eval_results(
        metrics_df, dates, eval, resampling, use_cv, config
    )
    return metrics_df, metrics_dict


def _preprocess_eval_df(evaluation_df: pd.DataFrame, use_cv: bool) -> pd.DataFrame:
    """Preprocesses evaluation dataframe.

    Parameters
    ----------
    evaluation_df : pd.DataFrame
        Evaluation dataframe.
    use_cv : bool
        Whether or note cross-validation is used.

    Returns
    -------
    pd.DataFrame
        Preprocessed evaluation dataframe.
    """
    if use_cv:
        df = evaluation_df.copy()
    else:
        df = add_time_groupers(evaluation_df)
    return df


def _compute_metrics(df: pd.DataFrame, eval: Dict[Any, Any]) -> pd.DataFrame:
    """Computes all metrics and gather them in a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Evaluation dataframe.
    eval : Dict
        Evaluation specifications.

    Returns
    -------
    pd.DataFrame
        Dataframe with all metrics at the desired granularity.
    """
    metrics = {"MAPE": MAPE, "SMAPE": SMAPE, "MSE": MSE, "RMSE": RMSE, "MAE": MAE}
    if eval["get_perf_on_agg_forecast"]:
        metrics_df = (
            df.groupby(eval["granularity"]).agg({"truth": "sum", "forecast": "sum"}).reset_index()
        )
        for m in eval["metrics"]:
            metrics_df[m] = metrics_df[["truth", "forecast"]].apply(
                lambda x: metrics[m](x.truth, x.forecast), axis=1
            )
    else:
        metrics_df = pd.DataFrame({eval["granularity"]: sorted(df[eval["granularity"]].unique())})
        for m in eval["metrics"]:
            metrics_df[m] = (
                df.groupby(eval["granularity"])[["truth", "forecast"]]
                .apply(lambda x: metrics[m](x.truth, x.forecast))
                .sort_index()
                .to_list()
            )
    return metrics_df


def _format_eval_results(
    metrics_df: pd.DataFrame,
    dates: Dict[Any, Any],
    eval: Dict[Any, Any],
    resampling: Dict[Any, Any],
    use_cv: bool,
    config: Dict[Any, Any],
) -> Tuple[pd.DataFrame, Dict[Any, Any]]:
    """Formats dataframe containing evaluation results and creates a dictionary containing the same information.

    Parameters
    ----------
    metrics_df : pd.DataFrame
        Dataframe with all metrics at the desired granularity.
    dates : Dict
        Dictionary containing all dates information.
    eval : Dict
        Evaluation specifications.
    resampling : Dict
        Resampling specifications.
    use_cv : bool
        Whether or note cross-validation is used.
    config : Dict
        Lib configuration dictionary.

    Returns
    -------
    pd.DataFrame
        Formatted dataframe with all metrics at the desired granularity.
    dict
        Dictionary with all metrics at the desired granularity.
    """
    if use_cv:
        metrics_df = __format_metrics_df_cv(metrics_df, dates, eval, resampling)
        metrics_dict = {m: metrics_df[[eval["granularity"], m]] for m in eval["metrics"]}
        metrics_df = __add_avg_std_metrics(metrics_df, eval)
    else:
        metrics_dict = {m: metrics_df[[eval["granularity"], m]] for m in eval["metrics"]}
        metrics_df = metrics_df[[eval["granularity"]] + eval["metrics"]].set_index(
            [eval["granularity"]]
        )
    metrics_df = __format_metrics_values(metrics_df, eval, config)
    return metrics_df, metrics_dict


def __format_metrics_values(
    metrics_df: pd.DataFrame, eval: Dict[Any, Any], config: Dict[Any, Any]
) -> pd.DataFrame:
    """Formats metrics values with the right number of decimals.

    Parameters
    ----------
    metrics_df : pd.DataFrame
        Dataframe with all metrics at the desired granularity.
    eval : Dict
        Evaluation specifications.
    config : Dict
        Lib configuration dictionary containing information about the number of decimals to keep.

    Returns
    -------
    pd.DataFrame
        Dataframe with all metrics formatted with the right number of decimals.
    """
    mapping_format = {k: "{:,." + str(v) + "f}" for k, v in config["metrics"]["digits"].items()}
    mapping_round = config["metrics"]["digits"].copy()
    for col in eval["metrics"]:
        metrics_df[col] = metrics_df[col].map(
            lambda x: mapping_format[col].format(round(x, mapping_round[col]))
        )
    return metrics_df


def __format_metrics_df_cv(
    metrics_df: pd.DataFrame,
    dates: Dict[Any, Any],
    eval: Dict[Any, Any],
    resampling: Dict[Any, Any],
) -> pd.DataFrame:
    """Formats dataframe containing evaluation metrics, in case cross-validation is used.

    Parameters
    ----------
    metrics_df : pd.DataFrame
        Dataframe with all metrics for each cross-validation fold.
    dates : Dict
        Dictionary containing cross-validation dates information.
    eval : Dict
        Evaluation specifications.
    resampling : Dict
        Resampling specifications.

    Returns
    -------
    pd.DataFrame
        Formatted dataframe with all metrics displayed for each cross-validation fold.
    """
    metrics_df = metrics_df.rename(columns={"cutoff": "Valid Start"})
    freq = resampling["freq"][-1]
    horizon = dates["folds_horizon"]
    if freq in ["s", "H"]:
        metrics_df["Valid End"] = (
            metrics_df["Valid Start"]
            .map(lambda x: x + timedelta(seconds=convert_into_nb_of_seconds(freq, horizon)))
            .astype(str)
        )
    else:
        metrics_df["Valid End"] = (
            metrics_df["Valid Start"]
            .map(lambda x: x + timedelta(days=convert_into_nb_of_days(freq, horizon)))
            .astype(str)
        )
    metrics_df["Valid Start"] = metrics_df["Valid Start"].astype(str)
    metrics_df = metrics_df.sort_values("Valid Start", ascending=False).reset_index(drop=True)
    metrics_df[eval["granularity"]] = [f"Fold {i}" for i in range(1, len(metrics_df) + 1)]
    return metrics_df


def __add_avg_std_metrics(metrics_df: pd.DataFrame, eval: Dict[Any, Any]) -> pd.DataFrame:
    """Adds rows for average and standard-deviation over each fold to dataframe containing evaluation metrics.

    Parameters
    ----------
    metrics_df : pd.DataFrame
        Dataframe with all metrics for each cross-validation fold.
    eval : Dict
        Evaluation specifications.

    Returns
    -------
    pd.DataFrame
        Formatted dataframe with two more rows (for average and standard-deviation of each metrics).
    """
    cols_index = [eval["granularity"], "Valid Start", "Valid End"]
    metrics_df = metrics_df[cols_index + eval["metrics"]].set_index(cols_index)
    metrics_df.loc[("Avg", "", "Average")] = metrics_df.mean(axis=0)
    metrics_df.loc[("Std", "", "+/-")] = metrics_df.std(axis=0)
    metrics_df = metrics_df.reset_index().set_index(eval["granularity"])
    return metrics_df



from typing import Any, Dict, List

import re
from datetime import datetime, timedelta

import pandas as pd
import streamlit as st
from streamlit_prophet.lib.dataprep.clean import clean_future_df
from streamlit_prophet.lib.dataprep.format import prepare_future_df
from streamlit_prophet.lib.utils.mapping import convert_into_nb_of_days, convert_into_nb_of_seconds


def get_train_val_sets(
    df: pd.DataFrame, dates: Dict[Any, Any], config: Dict[Any, Any], datasets: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Adds training and validation dataframes in datasets dictionary's values.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe containing both training and validation samples.
    dates : Dict
        Dictionary containing training and validation dates information.
    config : Dict
        Lib configuration dictionary.
    datasets : Dict
        Empty dictionary.

    Returns
    -------
    dict
        The datasets dictionary containing training and validation dataframes.
    """
    train = df.query(
        f'ds >= "{dates["train_start_date"]}" & ds <= "{dates["train_end_date"]}"'
    ).copy()
    val = df.query(f'ds >= "{dates["val_start_date"]}" & ds <= "{dates["val_end_date"]}"').copy()
    datasets["train"], datasets["val"] = train, val
    raise_error_train_val_dates(val, train, config, dates)
    print_train_val_dates(val, train)
    return datasets


def print_train_val_dates(val: pd.DataFrame, train: pd.DataFrame) -> None:
    """Displays a message in streamlit dashboard with training and validation dates.

    Parameters
    ----------
    val : pd.DataFrame
        Dataframe containing validation data.
    train : pd.DataFrame
        Dataframe containing training data.
    """
    st.success(
        f"""Train:              \n"""
        f"""[ {train.ds.min().strftime('%Y/%m/%d')} - {train.ds.max().strftime('%Y/%m/%d')} ]              \n"""
        f"""Valid:              \n"""
        f"""[ {val.ds.min().strftime('%Y/%m/%d')} - {val.ds.max().strftime('%Y/%m/%d')} ]              \n"""
        f"""({round((len(val) / float(len(train) + len(val)) * 100))}% of data used for validation)"""
    )


def raise_error_train_val_dates(
    val: pd.DataFrame, train: pd.DataFrame, config: Dict[Any, Any], dates: Dict[Any, Any]
) -> None:
    """Displays a message in streamlit dashboard and stops it if training and/or validation dates are incorrect.

    Parameters
    ----------
    val : pd.DataFrame
        Dataframe containing validation data.
    train : pd.DataFrame
        Dataframe containing training data.
    config : Dict
        Lib configuration dictionary where rules for training and validation dates are given.
    dates : Dict
        Dictionary containing training and validation dates information.
    """
    threshold_train = config["validity"]["min_data_points_train"]
    threshold_val = config["validity"]["min_data_points_val"]
    if dates["train_end_date"] >= dates["val_start_date"]:
        st.error(f"Training end date should be before validation start date.")
        st.stop()
    if dates["val_start_date"] >= dates["val_end_date"]:
        st.error(f"Validation start date should be before validation end date.")
        st.stop()
    if dates["train_start_date"] >= dates["train_end_date"]:
        st.error(f"Training start date should be before training end date.")
        st.stop()
    if len(val) <= threshold_val:
        st.error(
            f"There are less than {threshold_val + 1} data points in validation set ({len(val)}), "
            f"please expand validation period or change the dataset frequency. "
            f"If you wish to train a model on the whole dataset and forecast on future dates, "
            f"please go to the 'Forecast' section at the bottom of the sidebar."
        )
        st.stop()
    if len(train) <= threshold_train:
        st.error(
            f"There are less than {threshold_train + 1} data points in training set ({len(train)}), "
            f"please expand training period or change the dataset frequency."
        )
        st.stop()


def get_train_set(
    df: pd.DataFrame, dates: Dict[Any, Any], datasets: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Adds training dataframe in datasets dictionary's values.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe containing both training and validation samples.
    dates : Dict
        Dictionary containing training dates information.
    datasets : Dict
        Empty dictionary.

    Returns
    -------
    dict
        The datasets dictionary containing training dataframe.
    """
    train = df.query(
        f'ds >= "{dates["train_start_date"]}" & ds <= "{dates["train_end_date"]}"'
    ).copy()
    datasets["train"] = train
    return datasets


def make_eval_df(datasets: Dict[Any, Any]) -> Dict[Any, Any]:
    """Adds evaluation dataframe in datasets dictionary's values.

    Parameters
    ----------
    datasets : Dict
        Dictionary containing training and validation dataframes.

    Returns
    -------
    dict
        The datasets dictionary containing evaluation dataframe.
    """
    eval = pd.concat([datasets["train"], datasets["val"]], axis=0)
    eval = eval.drop("y", axis=1)
    datasets["eval"] = eval
    return datasets


def make_future_df(
    dates: Dict[Any, Any],
    df: pd.DataFrame,
    datasets: Dict[Any, Any],
    cleaning: Dict[Any, Any],
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
    load_options: Dict[Any, Any],
    config: Dict[Any, Any],
    resampling: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Adds future dataframe in datasets dictionary's values.

    Parameters
    ----------
    dates : Dict
        Dictionary containing future forecasting dates information.
    df : pd.DataFrame
        Full input dataframe, after cleaning, filtering and resampling.
    datasets : Dict
        Dictionary storing all dataframes.
    cleaning : Dict
        Cleaning specifications to apply to future dataframe.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.
    load_options : Dict
        Loading options selected by user.
    config : Dict
        Lib configuration dictionary.
    resampling : Dict
        Resampling specifications.

    Returns
    -------
    dict
        The datasets dictionary containing future dataframe.
    """
    datasets["full"] = df.copy()
    future, datasets = prepare_future_df(
        datasets, dates, date_col, target_col, dimensions, load_options, config, resampling
    )
    future = clean_future_df(future, cleaning)
    datasets["future"] = future
    return datasets


def get_train_end_date_default_value(
    df: pd.DataFrame,
    dates: Dict[Any, Any],
    resampling: Dict[Any, Any],
    config: Dict[Any, Any],
    use_cv: bool,
) -> pd.Timestamp:
    """Calculates training end date default value in streamlit dashboard.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe containing all observations.
    dates : Dict
        Dictionary containing training start date information.
    resampling : Dict
        Dictionary containing dataset frequency information.
    config : Dict
        Lib configuration dictionary containing validation period length information.
    use_cv : bool
        Whether or not cross-validation is used.

    Returns
    -------
    pd.Timestamp
        Training end date default value in streamlit dashboard.
    """
    if use_cv:
        default_end = df.ds.max()
    else:
        total_nb_days = (df.ds.max().date() - dates["train_start_date"]).days
        freq = resampling["freq"][-1]
        default_horizon = convert_into_nb_of_days(freq, config["horizon"][freq])
        default_end = df.ds.max() - timedelta(days=min(default_horizon, total_nb_days - 1))
    return default_end


def get_cv_cutoffs(dates: Dict[Any, Any], freq: str) -> List[Any]:
    """Generates the list of cross-validation cutoff dates.

    Parameters
    ----------
    dates : Dict
        Dictionary containing cross-validation dates information (number of folds, horizon, end date).
    freq : str
        Dataset frequency.

    Returns
    -------
    list
        List of all cross-validation cutoff dates.
    """
    horizon, end, n_folds = dates["folds_horizon"], dates["train_end_date"], dates["n_folds"]
    if freq in ["s", "H"]:
        end = datetime.combine(end, datetime.min.time())
        cutoffs = [
            pd.to_datetime(
                end - timedelta(seconds=(i + 1) * convert_into_nb_of_seconds(freq, horizon))
            )
            for i in range(n_folds)
        ]
    else:
        cutoffs = [
            pd.to_datetime(end - timedelta(days=(i + 1) * convert_into_nb_of_days(freq, horizon)))
            for i in range(n_folds)
        ]
    return cutoffs


def get_max_possible_cv_horizon(dates: Dict[Any, Any], resampling: Dict[Any, Any]) -> int:
    """Calculates maximum possible cross-validation horizon value in streamlit dashboard.

    Parameters
    ----------
    dates : Dict
        Dictionary containing training date information and number of cross-validation folds.
    resampling : Dict
        Dictionary containing dataset frequency information.

    Returns
    -------
    int
        Maximum possible cross-validation horizon value in streamlit dashboard.
    """
    freq = resampling["freq"][-1]
    if freq in ["s", "H"]:
        nb_seconds_training = (dates["train_end_date"] - dates["train_start_date"]).days * (
            24 * 60 * 60
        )
        max_horizon = (nb_seconds_training // convert_into_nb_of_seconds(freq, 1)) // dates[
            "n_folds"
        ]
    else:
        nb_days_training = (dates["train_end_date"] - dates["train_start_date"]).days
        max_horizon = (nb_days_training // convert_into_nb_of_days(freq, 1)) // dates["n_folds"]
    return int(max_horizon)


def print_cv_folds_dates(dates: Dict[Any, Any], freq: str) -> None:
    """Displays a message in streamlit dashboard with cross-validation folds' dates.

    Parameters
    ----------
    dates : Dict
        Dictionary containing cross-validation dates information.
    freq : str
        Dataset frequency.
    """
    horizon, cutoffs_text = dates["folds_horizon"], []
    for i, cutoff in enumerate(dates["cutoffs"]):
        cutoffs_text.append(f"""Fold {i + 1}:           """)
        if freq in ["s", "H"]:
            cutoffs_text.append(
                f"""Train:              \n"""
                f"""[ {dates['train_start_date'].strftime('%Y/%m/%d %H:%M:%S')} - """
                f"""{cutoff.strftime('%Y/%m/%d %H:%M:%S')} ]              """
            )
            cutoffs_text.append(
                f"""Valid:              \n"""
                f"""] {cutoff.strftime('%Y/%m/%d %H:%M:%S')} - """
                f"""{(cutoff + timedelta(seconds=convert_into_nb_of_seconds(freq, horizon)))
                                .strftime('%Y/%m/%d %H:%M:%S')} ]              \n"""
            )
        else:
            cutoffs_text.append(
                f"""Train:              \n"""
                f"""[ {dates['train_start_date'].strftime('%Y/%m/%d')} - """
                f"""{cutoff.strftime('%Y/%m/%d')} ]              """
            )
            cutoffs_text.append(
                f"""Valid:              \n"""
                f"""] {cutoff.strftime('%Y/%m/%d')} - """
                f"""{(cutoff + timedelta(days=convert_into_nb_of_days(freq, horizon)))
                    .strftime('%Y/%m/%d')} ]              \n"""
            )
        cutoffs_text.append("")
    st.success("\n".join(cutoffs_text))


def raise_error_cv_dates(
    dates: Dict[Any, Any], resampling: Dict[Any, Any], config: Dict[Any, Any]
) -> None:
    """Displays a message in streamlit dashboard and stops it if cross-validation dates are incorrect.

    Parameters
    ----------
    dates : Dict
        Dictionary containing cross-validation dates information.
    resampling : Dict
        Dictionary containing dataset frequency information.
    config : Dict
        Lib configuration dictionary where rules for cross-validation dates are given.
    """
    threshold_train = config["validity"]["min_data_points_train"]
    threshold_val = config["validity"]["min_data_points_val"]
    freq = resampling["freq"]
    regex = re.findall(r"\d+", resampling["freq"])
    freq_int = int(regex[0]) if len(regex) > 0 else 1
    n_data_points_val = dates["folds_horizon"] // freq_int
    n_data_points_train = len(
        pd.date_range(start=dates["train_start_date"], end=min(dates["cutoffs"]), freq=freq)
    )
    if n_data_points_val <= threshold_val:
        st.error(
            f"Some folds' valid sets have less than {threshold_val + 1} data points ({n_data_points_val}), "
            f"please increase folds' horizon or change the dataset frequency or expand CV period."
        )
        st.stop()
    elif n_data_points_train <= threshold_train:
        st.error(
            f"Some folds' train sets have less than {threshold_train + 1} data points ({n_data_points_train}), "
            f"please increase folds' horizon or change the dataset frequency or expand CV period."
        )
        st.stop()


def print_forecast_dates(dates: Dict[Any, Any], resampling: Dict[Any, Any]) -> None:
    """Displays a message in streamlit dashboard with future forecast dates.

    Parameters
    ----------
    dates : Dict
        Dictionary containing future forecast dates information.
    resampling : str
        Dictionary containing dataset frequency information.
    """
    if resampling["freq"][-1] in ["s", "H"]:
        st.success(
            f"""Forecast:              \n"""
            f"""{dates['forecast_start_date'].strftime('%Y/%m/%d %H:%M:%S')} -
                {dates['forecast_end_date'].strftime('%Y/%m/%d %H:%M:%S')}"""
        )
    else:
        st.success(
            f"""Forecast:              \n"""
            f"""{dates['forecast_start_date'].strftime('%Y/%m/%d')} -
                {dates['forecast_end_date'].strftime('%Y/%m/%d')}"""
        )



from typing import Any, Dict, List, Optional, Tuple

import datetime
from collections import defaultdict
from datetime import timedelta

import pandas as pd
from fbprophet import Prophet
from streamlit_prophet.lib.utils.mapping import convert_into_nb_of_days, convert_into_nb_of_seconds


def get_forecast_components(
    model: Prophet, forecast_df: pd.DataFrame, include_yhat: bool = False
) -> pd.DataFrame:
    """Returns a dataframe with only the relevant components to sum to get the prediction.

    Parameters
    ----------
    model : Prophet
        Fitted model.
    forecast_df : pd.DataFrame
        Forecast dataframe returned by Prophet model when predicting on evaluation dataset.
    include_yhat : bool
        Whether or nto to include yhat in columns.

    Returns
    -------
    pd.DataFrame
        Dataframe with only the relevant components to sum to get the prediction.
    """
    fcst = forecast_df.copy()
    components_col_names = get_forecast_components_col_names(fcst) + ["ds"]
    if include_yhat:
        components_col_names = components_col_names + ["yhat"]
    components = fcst[components_col_names]
    for col in components_col_names:
        if col in model.component_modes["multiplicative"]:
            components[col] *= components["trend"]

    components = components.set_index("ds")
    components_mapping = get_components_mapping(components, model, cols_to_drop=["holidays"])
    components = group_components(components, components_mapping)
    return components


def get_forecast_components_col_names(forecast_df: pd.DataFrame) -> List[Any]:
    """Returns the list of columns to keep in forecast dataframe to get all components without upper/lower bounds.

    Parameters
    ----------
    forecast_df : pd.DataFrame
        Forecast dataframe returned by Prophet model when predicting on evaluation dataset.

    Returns
    -------
    list
        List of columns to keep in forecast dataframe to get all components without upper/lower bounds.
    """
    components_col = [
        col.replace("_lower", "")
        for col in forecast_df.columns
        if "lower" in col
        and "yhat" not in col
        and "multiplicative" not in col
        and "additive" not in col
    ]
    return components_col


def get_components_mapping(
    components: pd.DataFrame, model: Prophet, cols_to_drop: Optional[List[str]] = None
) -> Dict[str, List[Any]]:
    """Compute a dict with value: list of columns to sum under key: new column name
    All columns in components will have as key either:
    - their own name (col: [col]) -> to be left as is
    - a new name (new_col_name: [col_1, col_2]) -> to be grouped
    - '_to_drop_' ('_to_drop_': [col_to_drop_1, col_to_drop_2]) -> to be dropped

    Parameters
    ----------
    components: pd.DataFrame
        Forecast output from Prophet model with pre-selected components only
    model: Prophet
        Fitted prophet model
    cols_to_drop: List[str]
        List of columns to be dropped. Other columns will be either kept or left as is

    Returns
    -------
    Dict[str, list]
        dict with value: list of columns to sum under key: new column name
    """
    if cols_to_drop is None:
        cols_to_drop = []

    components_mapping = defaultdict(list)
    for col in components.columns:
        if (
            model.train_holiday_names is not None and col in model.train_holiday_names.values
        ):  # group
            if col.startswith("School holiday"):
                components_mapping["School holidays"].append(col)
            elif col.startswith("Lockdown"):
                components_mapping["Lockdown events"].append(col)
            else:
                components_mapping["Public holidays"].append(col)
        elif col in cols_to_drop:  # drop
            components_mapping["_to_drop_"].append(col)
        else:
            components_mapping[col].append(col)  # left as is
    return components_mapping


def group_components(
    components: pd.DataFrame, components_mapping: Dict[str, List[Any]]
) -> pd.DataFrame:
    """Group components based on components_mapping in a copy of the components df

    Parameters
    ----------
    components: pd.DataFrame
        Forecast output from Prophet model with pre-selected components only
    components_mapping: Dict[str, list]
        dict with value: list of columns to sum under key: new column name. \
Can include a '_to_drop_' item to mark columns to be dropped. \
A column not in any of the mapping values will be dropped.

    Returns
    -------
    pd.DataFrame
        Dataframe with components either left as is, summed or dropped, based on provided mapping
    """
    grouped_components = pd.DataFrame(index=components.index)
    for new_col_name, grouped_cols in components_mapping.items():
        if new_col_name != "_to_drop_":
            grouped_components[new_col_name] = components[grouped_cols].sum(axis=1)
    return grouped_components


def get_df_cv_with_hist(
    forecasts: Dict[Any, Any], datasets: Dict[Any, Any], models: Dict[Any, Any]
) -> pd.DataFrame:
    """Adds training rows not included in CV validation folds to the dataframe containing cross-validation results.

    Parameters
    ----------
    forecasts : Dict
        Dictionary containing the dataframe with cross-validation results.
    datasets : Dict
        Dictionary containing training dataframe.
    models : Dict
        Dictionary containing the model fitted for evaluation.

    Returns
    -------
    pd.DataFrame
        Dataframe containing CV results and predictions on training data not included in CV validation folds.
    """
    df_cv = forecasts["cv"].drop(["cutoff"], axis=1)
    df_past = models["eval"].predict(
        datasets["train"].loc[datasets["train"]["ds"] < df_cv.ds.min()].drop("y", axis=1)
    )
    common_cols = ["ds", "yhat", "yhat_lower", "yhat_upper"]
    df_past = df_past[common_cols + list(set(df_past.columns) - set(common_cols))]
    df_cv = pd.concat([df_cv, df_past], axis=0).sort_values("ds").reset_index(drop=True)
    return df_cv


def get_cv_dates_dict(dates: Dict[Any, Any], resampling: Dict[Any, Any]) -> Dict[Any, Any]:
    """Returns a dictionary whose keys are CV folds and values are dictionaries with each fold's train/valid dates.

    Parameters
    ----------
    dates : Dict
        Dictionary containing cross-validation dates information.
    resampling : Dict
        Dictionary containing dataset frequency information.

    Returns
    -------
    dict
        Dictionary containing training and validation dates of each cross-validation fold.
    """
    freq = resampling["freq"][-1]
    train_start = dates["train_start_date"]
    horizon = dates["folds_horizon"]
    cv_dates: Dict[Any, Any] = dict()
    for i, cutoff in sorted(enumerate(dates["cutoffs"]), reverse=True):
        cv_dates[f"Fold {i + 1}"] = dict()
        cv_dates[f"Fold {i + 1}"]["train_start"] = train_start
        cv_dates[f"Fold {i + 1}"]["val_start"] = cutoff
        cv_dates[f"Fold {i + 1}"]["train_end"] = cutoff
        if freq in ["s", "H"]:
            cv_dates[f"Fold {i + 1}"]["val_end"] = cutoff + timedelta(
                seconds=convert_into_nb_of_seconds(freq, horizon)
            )
        else:
            cv_dates[f"Fold {i + 1}"]["val_end"] = cutoff + timedelta(
                days=convert_into_nb_of_days(freq, horizon)
            )
    return cv_dates


def get_hover_template_cv(
    cv_dates: Dict[Any, Any], resampling: Dict[Any, Any]
) -> Tuple[pd.DataFrame, str]:
    """Returns a dataframe and a dictionary that will be used to show CV folds on a plotly bar plot.

    Parameters
    ----------
    cv_dates : Dict
        Dictionary containing training and validation dates of each cross-validation fold.
    resampling : Dict
        Dictionary containing dataset frequency information.

    Returns
    -------
    pd.DataFrame
        Dataframe that will be used to plot cross-validation folds with plotly.
    str
        Hover template that will be used to show cross-validation folds dates on a plotly viz.
    """
    hover_data = pd.DataFrame(cv_dates).T
    if resampling["freq"][-1] in ["s", "H"]:
        hover_data = hover_data.applymap(lambda x: x.strftime("%Y/%m/%d %H:%M:%S"))
    else:
        hover_data = hover_data.applymap(lambda x: x.strftime("%Y/%m/%d"))
    hover_template = "<br>".join(
        [
            "%{y}",
            "Training start date: %{text[0]}",
            "Training end date: %{text[2]}",
            "Validation start date: %{text[1]}",
            "Validation end date: %{text[3]}",
        ]
    )
    return hover_data, hover_template


def prepare_waterfall(
    components: pd.DataFrame, start_date: datetime.date, end_date: datetime.date
) -> pd.DataFrame:
    """Returns a dataframe with only the relevant components to sum to get the prediction.

    Parameters
    ----------
    components : pd.DataFrame
        Dataframe with relevant components
    start_date : datetime.date
        Start date for components computation.
    end_date : datetime.date
        End date for components computation.

    Returns
    -------
    pd.DataFrame
        Dataframe with only the relevant data to plot the waterfall chart.
    """
    waterfall = components.loc[
        (components["ds"] >= pd.to_datetime(start_date))
        & (components["ds"] < pd.to_datetime(end_date))
    ]
    waterfall = waterfall.mean(axis=0, numeric_only=True)
    waterfall = waterfall[waterfall != 0]
    return waterfall



from typing import Any, Dict, List

import pandas as pd
import streamlit as st
from streamlit_prophet.lib.utils.holidays import lockdown_format_func
from streamlit_prophet.lib.utils.mapping import (
    COUNTRY_NAMES_MAPPING,
    COVID_LOCKDOWN_DATES_MAPPING,
    SCHOOL_HOLIDAYS_FUNC_MAPPING,
)


def input_seasonality_params(
    config: Dict[Any, Any],
    params: Dict[Any, Any],
    resampling: Dict[Any, Any],
    readme: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Lets the user enter seasonality parameters.

    Parameters
    ----------
    params : Dict
        Model parameters.
    config : Dict
        Lib config dictionary containing information about default parameters.
    resampling : Dict
        Dictionary containing dataset frequency information.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Model parameters with seasonality parameters added.
    """
    default_params = config["model"]
    seasonalities: Dict[str, Dict[Any, Any]] = {
        "yearly": {"period": 365.25, "prophet_param": None},
        "monthly": {"period": 30.5, "prophet_param": None},
        "weekly": {"period": 7, "prophet_param": None},
    }
    if resampling["freq"][-1] in ["s", "H"]:
        seasonalities["daily"] = {"period": 1, "prophet_param": None}
    for seasonality, values in seasonalities.items():

        values["prophet_param"] = st.selectbox(
            f"{seasonality.capitalize()} seasonality",
            ["auto", False, "custom"] if seasonality[0] in ["y", "w", "d"] else [False, "custom"],
            help=readme["tooltips"]["seasonality"],
        )
        if values["prophet_param"] == "custom":
            values["prophet_param"] = False
            values["custom_param"] = {
                "name": seasonality,
                "period": values["period"],
                "mode": st.selectbox(
                    f"Seasonality mode for {seasonality} seasonality",
                    default_params["seasonality_mode"],
                    help=readme["tooltips"]["seasonality_mode"],
                ),
                "fourier_order": st.number_input(
                    f"Fourier order for {seasonality} seasonality",
                    value=15,
                    help=readme["tooltips"]["seasonality_fourier"],
                ),
                "prior_scale": st.number_input(
                    f"Prior scale for {seasonality} seasonality",
                    value=10,
                    help=readme["tooltips"]["seasonality_prior_scale"],
                ),
            }
    add_custom_seasonality = st.checkbox(
        "Add a custom seasonality", value=False, help=readme["tooltips"]["add_custom_seasonality"]
    )
    if add_custom_seasonality:
        custom_seasonality: Dict[Any, Any] = dict()
        custom_seasonality["custom_param"] = dict()
        custom_seasonality["custom_param"]["name"] = st.text_input(
            "Name", value="custom_seasonality", help=readme["tooltips"]["seasonality_name"]
        )
        custom_seasonality["custom_param"]["period"] = st.number_input(
            "Period (in days)", value=10, help=readme["tooltips"]["seasonality_period"]
        )
        custom_seasonality["custom_param"]["mode"] = st.selectbox(
            f"Mode", default_params["seasonality_mode"], help=readme["tooltips"]["seasonality_mode"]
        )
        custom_seasonality["custom_param"]["fourier_order"] = st.number_input(
            f"Fourier order", value=15, help=readme["tooltips"]["seasonality_fourier"]
        )
        custom_seasonality["custom_param"]["prior_scale"] = st.number_input(
            f"Prior scale", value=10, help=readme["tooltips"]["seasonality_prior_scale"]
        )
        seasonalities[custom_seasonality["custom_param"]["name"]] = custom_seasonality
    params["seasonalities"] = seasonalities
    return params


def input_prior_scale_params(config: Dict[Any, Any], readme: Dict[Any, Any]) -> Dict[Any, Any]:
    """Lets the user enter prior scale parameters.

    Parameters
    ----------
    config : Dict
        Lib config dictionary containing information about default parameters.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Model prior scale parameters.
    """
    params = dict()
    default_params = config["model"]
    changepoint_prior_scale = st.number_input(
        "changepoint_prior_scale",
        value=default_params["changepoint_prior_scale"],
        format="%.3f",
        help=readme["tooltips"]["changepoint_prior_scale"],
    )
    seasonality_prior_scale = st.number_input(
        "seasonality_prior_scale",
        value=default_params["seasonality_prior_scale"],
        help=readme["tooltips"]["seasonality_prior_scale"],
    )
    holidays_prior_scale = st.number_input(
        "holidays_prior_scale",
        value=default_params["holidays_prior_scale"],
        help=readme["tooltips"]["holidays_prior_scale"],
    )
    params["prior_scale"] = {
        "seasonality_prior_scale": seasonality_prior_scale,
        "holidays_prior_scale": holidays_prior_scale,
        "changepoint_prior_scale": changepoint_prior_scale,
    }
    return params


def input_other_params(
    config: Dict[Any, Any], params: Dict[Any, Any], readme: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user enter other parameters (growth, changepoints_range, n_changepoints).

    Parameters
    ----------
    config : Dict
        Lib config dictionary containing information about default parameters.
    params : Dict
        Model parameters.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Model parameters with other parameters added.
    """
    default_params = config["model"]
    changepoint_range = st.number_input(
        "changepoint_range",
        value=default_params["changepoint_range"],
        max_value=1.0,
        min_value=0.0,
        format="%.2f",
        help=readme["tooltips"]["changepoint_range"],
    )
    growth = st.selectbox("growth", default_params["growth"], help=readme["tooltips"]["growth"])
    params["other"] = {
        "growth": growth,
        "changepoint_range": changepoint_range,
    }
    if growth == "logistic":
        cap = st.number_input(
            "cap",
            value=default_params["cap"],
            format="%.1f",
            help=readme["tooltips"]["cap"],
        )
        floor = st.number_input(
            "floor",
            value=default_params["floor"],
            format="%.1f",
            max_value=cap,
            help=readme["tooltips"]["floor"],
        )
        params["saturation"] = {
            "cap": cap,
            "floor": floor,
        }
    return params


def input_holidays_params(
    params: Dict[Any, Any], readme: Dict[Any, Any], config: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user enter holidays parameters.

    Parameters
    ----------
    params : Dict
        Model parameters.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    config : Dict
        Dictionary where user can provide the list of countries whose holidays will be included.

    Returns
    -------
    dict
        Model parameters with holidays parameters added.
    """
    countries = list(COUNTRY_NAMES_MAPPING.keys())
    default_country = config["model"]["holidays_country"]
    country = st.selectbox(
        label="Select a country",
        options=countries,
        index=countries.index(default_country),
        format_func=lambda x: COUNTRY_NAMES_MAPPING[x],
        help=readme["tooltips"]["holidays_country"],
    )

    public_holidays = st.checkbox(
        label="Public holidays",
        value=config["model"]["public_holidays"],
        help=readme["tooltips"]["public_holidays"],
    )

    school_holidays = False
    if country in SCHOOL_HOLIDAYS_FUNC_MAPPING.keys():
        school_holidays = st.checkbox(
            label="School holidays",
            value=config["model"]["school_holidays"],
            help=readme["tooltips"]["school_holidays"],
        )

    lockdowns = []
    if country in COVID_LOCKDOWN_DATES_MAPPING.keys():
        lockdown_options = list(range(len(COVID_LOCKDOWN_DATES_MAPPING[country])))
        lockdowns = st.multiselect(
            label="Lockdown events",
            options=lockdown_options,
            default=config["model"]["lockdown_events"],
            format_func=lockdown_format_func,
            help=readme["tooltips"]["lockdown_events"],
        )

    params["holidays"] = {
        "country": country,
        "public_holidays": public_holidays,
        "school_holidays": school_holidays,
        "lockdown_events": lockdowns,
    }
    return params


def input_regressors(
    df: pd.DataFrame, config: Dict[Any, Any], params: Dict[Any, Any], readme: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user select regressors.

    Parameters
    ----------
    df : Dict
        Prepared dataset (after filtering, resampling, cleaning).
    config : Dict
        Lib config dictionary containing information about default parameters.
    params : Dict
        Model parameters.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Model parameters with regressors information added.
    """
    regressors: Dict[Any, Any] = dict()
    default_params = config["model"]
    all_cols = set(df.columns) - {"ds", "y"}
    mask = df[all_cols].isnull().sum() == 0
    eligible_cols = sorted(list(mask[mask].index))
    _print_removed_regressors(sorted(set(all_cols) - set(eligible_cols)))
    if len(eligible_cols) > 0:
        if st.checkbox(
            "Add all detected regressors",
            value=False,
            help=readme["tooltips"]["add_all_regressors"],
        ):
            default_regressors = list(eligible_cols)
        else:
            default_regressors = []
        config_regressors = config["columns"]["regressors"]
        if config_regressors not in ["false", False]:
            if len(set(config_regressors).intersection(set(eligible_cols))) != len(
                config_regressors
            ):
                st.error(
                    f"Selected regressors are not in the dataset columns, "
                    f"please provide a list of valid columns for regressors in the config file."
                )
                st.stop()
        regressor_cols = st.multiselect(
            "Select external regressors if any",
            list(eligible_cols),
            default=default_regressors
            if config_regressors in ["false", False]
            else config_regressors,
            help=readme["tooltips"]["select_regressors"],
        )
        for col in regressor_cols:
            regressors[col] = dict()
            regressors[col]["prior_scale"] = st.number_input(
                f"Prior scale for {col}",
                value=default_params["regressors_prior_scale"],
                help=readme["tooltips"]["regressor_prior_scale"],
            )
    else:
        st.write("There are no regressors in your dataset.")
    params["regressors"] = regressors
    return params


def _print_removed_regressors(nan_cols: List[Any]) -> None:
    """Displays a message in streamlit dashboard if the input list is not empty.

    Parameters
    ----------
    nan_cols : list
        List of columns that have been removed because they contain null values.
    """
    L = len(nan_cols)
    if L > 0:
        st.error(
            f'The following column{"s" if L > 1 else ""} cannot be taken as regressor because '
            f'{"they contain" if L > 1 else "it contains"} null values: {", ".join(nan_cols)}'
        )



from typing import Any, Dict

import pandas as pd


def get_evaluation_df(
    datasets: Dict[Any, Any],
    forecasts: Dict[Any, Any],
    dates: Dict[Any, Any],
    eval: Dict[Any, Any],
    use_cv: bool,
) -> pd.DataFrame:
    """Generates a dataframe that will be used for evaluation.

    Parameters
    ----------
    datasets : Dict
        Dictionary containing evaluation dataframe.
    forecasts : Dict
        Dictionary where all forecasts are stored.
    dates : Dict
        Dictionary containing all dates information.
    eval : Dict
        Evaluation specifications.
    use_cv : bool
        Whether or not cross-validation is used.

    Returns
    -------
    pd.DataFrame
        Evaluation dataframe.
    """
    if use_cv:
        evaluation_df = forecasts["cv"].rename(columns={"y": "truth", "yhat": "forecast"})
        mapping = {
            cutoff: f"Fold {i + 1}"
            for i, cutoff in enumerate(sorted(evaluation_df["cutoff"].unique(), reverse=True))
        }
        evaluation_df["Fold"] = evaluation_df["cutoff"].map(mapping)
        evaluation_df = evaluation_df.sort_values("ds")
    else:
        evaluation_df = pd.DataFrame()
        if eval["set"] == "Validation":
            evaluation_df["ds"] = datasets["val"].ds.copy()
            evaluation_df["truth"] = list(datasets["val"].y)
            evaluation_df["forecast"] = list(
                forecasts["eval"]
                .query(f'ds >= "{dates["val_start_date"]}" & ' f'ds <= "{dates["val_end_date"]}"')
                .yhat
            )
        elif eval["set"] == "Training":
            evaluation_df["ds"] = datasets["train"].ds.copy()
            evaluation_df["truth"] = list(datasets["train"].y)
            evaluation_df["forecast"] = list(
                forecasts["eval"]
                .query(
                    f'ds >= "{dates["train_start_date"]}" & ' f'ds <= "{dates["train_end_date"]}"'
                )
                .yhat
            )
    return evaluation_df


def add_time_groupers(evaluation_df: pd.DataFrame) -> pd.DataFrame:
    """Adds columns with time information (day, week, quarter, year) to evaluation dataframe.

    Parameters
    ----------
    evaluation_df : pd.DataFrame
        Dictionary containing evaluation dataframe.

    Returns
    -------
    pd.DataFrame
        Evaluation dataframe with additional time information columns.
    """
    df = evaluation_df.copy()
    df["Global"] = "Global"
    df["Daily"] = df["ds"].astype(str).map(lambda x: x[0:10])
    df["Day of Week"] = (
        df["ds"].dt.dayofweek.map(lambda x: x + 1).astype(str) + ". " + df["ds"].dt.day_name()
    )
    df["Weekly"] = (
        df["ds"].dt.year.astype(str)
        + " - W"
        + df["ds"].dt.isocalendar().week.astype(str).map(lambda x: "0" + x if len(x) < 2 else x)
    )
    df["Monthly"] = (
        df["ds"].dt.year.astype(str)
        + " - M"
        + df["ds"].dt.month.astype(str).map(lambda x: "0" + x if len(x) < 2 else x)
    )
    df["Quarterly"] = df["ds"].dt.year.astype(str) + " - Q" + df["ds"].dt.quarter.astype(str)
    df["Yearly"] = df["ds"].dt.year.astype(str)
    return df



from typing import Any, Dict, Tuple

import pandas as pd
import streamlit as st
from streamlit_prophet.lib.exposition.export import display_config_download_links
from streamlit_prophet.lib.utils.load import download_toy_dataset, load_custom_config, load_dataset


def input_dataset(
    config: Dict[Any, Any], readme: Dict[Any, Any], instructions: Dict[Any, Any]
) -> Tuple[pd.DataFrame, Dict[Any, Any], Dict[Any, Any], Dict[Any, Any]]:
    """Lets the user decide whether to upload a dataset or download a toy dataset.

    Parameters
    ----------
    config : Dict
        Lib config dictionary containing information about toy datasets (download links).
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    instructions : Dict
        Dictionary containing instructions to provide a custom config.

    Returns
    -------
    pd.DataFrame
        Selected dataset loaded into a dataframe.
    dict
        Loading options selected by user (upload or download, dataset name if download).
    dict
        Lib configuration dictionary.
    dict
        Dictionary containing all datasets.
    """
    load_options, datasets = dict(), dict()
    load_options["toy_dataset"] = st.checkbox(
        "Load a toy dataset", True, help=readme["tooltips"]["upload_choice"]
    )
    if load_options["toy_dataset"]:
        dataset_name = st.selectbox(
            "Select a toy dataset",
            options=list(config["datasets"].keys()),
            format_func=lambda x: config["datasets"][x]["name"],
            help=readme["tooltips"]["toy_dataset"],
        )
        df = download_toy_dataset(config["datasets"][dataset_name]["url"])
        load_options["dataset"] = dataset_name
        load_options["date_format"] = config["dataprep"]["date_format"]
        load_options["separator"] = ","
    else:
        file = st.file_uploader(
            "Upload a csv file", type="csv", help=readme["tooltips"]["dataset_upload"]
        )
        load_options["separator"] = st.selectbox(
            "What is the separator?", [",", ";", "|"], help=readme["tooltips"]["separator"]
        )
        load_options["date_format"] = st.text_input(
            "What is the date format?",
            config["dataprep"]["date_format"],
            help=readme["tooltips"]["date_format"],
        )
        if st.checkbox(
            "Upload my own config file", False, help=readme["tooltips"]["custom_config_choice"]
        ):
            with st.sidebar.beta_expander("Configuration", expanded=True):
                display_config_download_links(
                    config,
                    "config.toml",
                    "Template",
                    instructions,
                    "instructions.toml",
                    "Instructions",
                )
                config_file = st.file_uploader(
                    "Upload custom config", type="toml", help=readme["tooltips"]["custom_config"]
                )
                if config_file:
                    config = load_custom_config(config_file)
                else:
                    st.stop()
        if file:
            df = load_dataset(file, load_options)
        else:
            st.stop()
    datasets["uploaded"] = df.copy()
    return df, load_options, config, datasets


def input_columns(
    config: Dict[Any, Any], readme: Dict[Any, Any], df: pd.DataFrame, load_options: Dict[Any, Any]
) -> Tuple[str, str]:
    """Lets the user specify date and target column names.

    Parameters
    ----------
    config : Dict
        Lib config dictionary containing information about toy datasets (date and target column names).
    readme : Dict
        Dictionary containing tooltips to guide user's choices.
    df : pd.DataFrame
        Loaded dataset.
    load_options : Dict
        Loading options selected by user (upload or download, dataset name if download).

    Returns
    -------
    str
        Date column name.
    str
        Target column name.
    """
    if load_options["toy_dataset"]:
        date_col = st.selectbox(
            "Date column",
            [config["datasets"][load_options["dataset"]]["date"]],
            help=readme["tooltips"]["date_column"],
        )
        target_col = st.selectbox(
            "Target column",
            [config["datasets"][load_options["dataset"]]["target"]],
            help=readme["tooltips"]["target_column"],
        )
    else:
        date_col = st.selectbox(
            "Date column",
            sorted(df.columns)
            if config["columns"]["date"] in ["false", False]
            else [config["columns"]["date"]],
            help=readme["tooltips"]["date_column"],
        )
        target_col = st.selectbox(
            "Target column",
            sorted(set(df.columns) - {date_col})
            if config["columns"]["target"] in ["false", False]
            else [config["columns"]["target"]],
            help=readme["tooltips"]["target_column"],
        )
    return date_col, target_col


def input_future_regressors(
    datasets: Dict[Any, Any],
    dates: Dict[Any, Any],
    params: Dict[Any, Any],
    dimensions: Dict[Any, Any],
    load_options: Dict[Any, Any],
    date_col: str,
) -> pd.DataFrame:
    """Adds future regressors dataframe in datasets dictionary's values.

    Parameters
    ----------
    datasets : Dict
        Dictionary storing all dataframes.
    dates : Dict
        Dictionary containing future forecasting dates information.
    params : Dict
        Dictionary containing all model parameters and list of selected regressors.
    dimensions : Dict
        Dictionary containing dimensions information.
    load_options : Dict
        Loading options selected by user (including csv delimiter).
    date_col : str
        Name of date column.

    Returns
    -------
    dict
        The datasets dictionary containing future regressors dataframe.
    """
    if len(params["regressors"].keys()) > 0:
        regressors_col = list(params["regressors"].keys())
        start, end = dates["forecast_start_date"], dates["forecast_end_date"]
        tooltip = (
            f"Please upload a csv file with delimiter '{load_options['separator']}' "
            "and the same format as input dataset, ie with the following specifications: \n"
        )
        tooltip += (
            f"- Date column named `{date_col}`, going from **{start.strftime('%Y-%m-%d')}** "
            f"to **{end.strftime('%Y-%m-%d')}** at the same frequency as input dataset "
            f"and at format **{load_options['date_format']}**. \n"
        )
        dimensions_col = [col for col in dimensions.keys() if col != "agg"]
        if len(dimensions_col) > 0:
            if len(dimensions_col) > 1:
                tooltip += (
                    f"- Columns with the following names for dimensions: `{', '.join(dimensions_col[:-1])}, "
                    f"{dimensions_col[-1]}`. \n"
                )
            else:
                tooltip += f"- Dimension column named `{dimensions_col[0]}`. \n"
        if len(regressors_col) > 1:
            tooltip += (
                f"- Columns with the following names for regressors: `{', '.join(regressors_col[:-1])}, "
                f"{regressors_col[-1]}`."
            )
        else:
            tooltip += f"- Regressor column named `{regressors_col[0]}`."
        regressors_file = st.file_uploader(
            "Upload a csv file for regressors", type="csv", help=tooltip
        )
        if regressors_file:
            datasets["future_regressors"] = load_dataset(regressors_file, load_options)
    else:
        st.write("There are no regressors selected.")
    return datasets



from typing import Any, List


def reverse_list(L: List[Any], N: int) -> List[Any]:
    """Cuts the list after the N-th element and reverses its order.

    Parameters
    ----------
    L : list
        List to be reversed.
    N : int
        Index at which the list will be cut if it is smaller than the list length.

    Returns
    -------
    list
        Reversed list.
    """
    if N < len(L):
        L = L[:N]
    reversed_list = [L[len(L) - 1 - i] for i, x in enumerate(L)]
    return reversed_list



import typer
from rich.console import Console
from streamlit_prophet.app import deploy_streamlit

app = typer.Typer()
console = Console()


@app.command()
def dashboard() -> None:
    """Deploys the streamlit dashboard."""
    deploy_streamlit()



from typing import Any, Dict, Tuple

import numpy as np
import pandas as pd
import streamlit as st


def clean_df(df: pd.DataFrame, cleaning: Dict[Any, Any]) -> pd.DataFrame:
    """Cleans the input dataframe according to cleaning dict specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that has to be cleaned.
    cleaning : Dict
        Cleaning specifications.

    Returns
    -------
    pd.DataFrame
        Cleaned dataframe.
    """
    df = _remove_rows(df, cleaning)
    df = _log_transform(df, cleaning)
    return df


def clean_future_df(df: pd.DataFrame, cleaning: Dict[Any, Any]) -> pd.DataFrame:
    """Cleans the input dataframe according to cleaning dict specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that has to be cleaned.
    cleaning : Dict
        Cleaning specifications.

    Returns
    -------
    pd.DataFrame
        Cleaned dataframe.
    """
    df_clean = df.copy()  # To avoid CachedObjectMutationWarning
    df_clean["__to_remove"] = 0
    if cleaning["del_days"] is not None:
        df_clean["__to_remove"] = np.where(
            df_clean.ds.dt.dayofweek.isin(cleaning["del_days"]), 1, df_clean["__to_remove"]
        )
    df_clean = df_clean.query("__to_remove != 1")
    del df_clean["__to_remove"]
    return df_clean


@st.cache(suppress_st_warning=True)
def _log_transform(df: pd.DataFrame, cleaning: Dict[Any, Any]) -> pd.DataFrame:
    """Applies a log transform to the y column of input dataframe, if possible.
    Raises an error in streamlit dashboard if not possible.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that has to be cleaned.
    cleaning : Dict
        Cleaning specifications.

    Returns
    -------
    pd.DataFrame
        Cleaned dataframe.
    """
    df_clean = df.copy()  # To avoid CachedObjectMutationWarning
    if cleaning["log_transform"]:
        if df_clean.y.min() <= 0:
            st.error(
                "The target has values <= 0. Please remove negative and 0 values when applying log transform."
            )
            st.stop()
        else:
            df_clean["y"] = np.log(df_clean["y"])
    return df_clean


@st.cache()
def _remove_rows(df: pd.DataFrame, cleaning: Dict[Any, Any]) -> pd.DataFrame:
    """Removes some rows of the input dataframe according to cleaning dict specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that has to be cleaned.
    cleaning : Dict
        Cleaning specifications.

    Returns
    -------
    pd.DataFrame
        Cleaned dataframe.
    """
    df_clean = df.copy()  # To avoid CachedObjectMutationWarning
    df_clean["__to_remove"] = 0
    if cleaning["del_negative"]:
        df_clean["__to_remove"] = np.where(df_clean["y"] < 0, 1, df_clean["__to_remove"])
    if cleaning["del_days"] is not None:
        df_clean["__to_remove"] = np.where(
            df_clean.ds.dt.dayofweek.isin(cleaning["del_days"]), 1, df_clean["__to_remove"]
        )
    if cleaning["del_zeros"]:
        df_clean["__to_remove"] = np.where(df_clean["y"] == 0, 1, df_clean["__to_remove"])
    df_clean = df_clean.query("__to_remove != 1")
    del df_clean["__to_remove"]
    return df_clean


def exp_transform(
    datasets: Dict[Any, Any], forecasts: Dict[Any, Any]
) -> Tuple[Dict[Any, Any], Dict[Any, Any]]:
    """Applies an exp transform to the y column of dataframes which are values of input dictionaries.

    Parameters
    ----------
    datasets : Dict
        A dictionary whose values are dataframes used as an input to fit a Prophet model.
    forecasts : Dict
        A dictionary whose values are dataframes which are the output of a Prophet prediction.

    Returns
    -------
    dict
        The datasets dictionary with transformed values.
    dict
        The forecasts dictionary with transformed values.
    """
    for data in set(datasets.keys()):
        if "y" in datasets[data].columns:
            df_exp = datasets[data].copy()
            df_exp["y"] = np.exp(df_exp["y"])
            datasets[data] = df_exp.copy()
    for data in set(forecasts.keys()):
        if "yhat" in forecasts[data].columns:
            df_exp = forecasts[data].copy()
            df_exp["yhat"] = np.exp(df_exp["yhat"])
            forecasts[data] = df_exp.copy()
    return datasets, forecasts



import os
import sys

from streamlit import cli


def deploy_streamlit():
    sys.argv = [
        "streamlit",
        "run",
        f"{os.path.dirname(os.path.realpath(__file__))}/dashboard.py",
        "--server.port=8080",
        "--server.address=0.0.0.0",
    ]
    sys.exit(cli.main())



from typing import Any, Dict, List, Tuple

from streamlit_prophet.lib.utils.holidays import get_school_holidays_FR

COUNTRY_NAMES_MAPPING = {
    "FR": "France",
    "US": "United States",
    "UK": "United Kingdom",
    "CA": "Canada",
    "BR": "Brazil",
    "MX": "Mexico",
    "IN": "India",
    "CN": "China",
    "JP": "Japan",
    "DE": "Germany",
    "IT": "Italy",
    "RU": "Russia",
    "BE": "Belgium",
    "PT": "Portugal",
    "PL": "Poland",
}

COVID_LOCKDOWN_DATES_MAPPING = {
    "FR": [
        ("2020-03-17", "2020-05-11"),
        ("2020-10-30", "2020-12-15"),
        ("2021-03-20", "2021-05-03"),
    ]
}

SCHOOL_HOLIDAYS_FUNC_MAPPING = {
    "FR": get_school_holidays_FR,
}


def convert_into_nb_of_days(freq: str, horizon: int) -> int:
    """Converts a forecasting horizon in number of days.

    Parameters
    ----------
    freq : str
        Dataset frequency.
    horizon : int
        Forecasting horizon in dataset frequency units.

    Returns
    -------
    int
        Forecasting horizon in days.
    """
    mapping = {
        "s": horizon // (24 * 60 * 60),
        "H": horizon // 24,
        "D": horizon,
        "W": horizon * 7,
        "M": horizon * 30,
        "Q": horizon * 90,
        "Y": horizon * 365,
    }
    return mapping[freq]


def convert_into_nb_of_seconds(freq: str, horizon: int) -> int:
    """Converts a forecasting horizon in number of seconds.

    Parameters
    ----------
    freq : str
        Dataset frequency.
    horizon : int
        Forecasting horizon in dataset frequency units.

    Returns
    -------
    int
        Forecasting horizon in seconds.
    """
    mapping = {
        "s": horizon,
        "H": horizon * 60 * 60,
        "D": horizon * 60 * 60 * 24,
        "W": horizon * 60 * 60 * 24 * 7,
        "M": horizon * 60 * 60 * 24 * 30,
        "Q": horizon * 60 * 60 * 24 * 90,
        "Y": horizon * 60 * 60 * 24 * 365,
    }
    return mapping[freq]


def dayname_to_daynumber(days: List[Any]) -> List[Any]:
    """Converts a list of day names into a list of day numbers from 0 (Monday) to 6 (Sunday).

    Parameters
    ----------
    days : list
        Day names.

    Returns
    -------
    list
        Day numbers from 0 (Monday) to 6 (Sunday).
    """
    day_names = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
    mapping = {day: i for i, day in enumerate(day_names)}
    return [mapping[day] for day in days]


def mapping_country_names(countries: List[Any]) -> Tuple[Dict[Any, Any], List[Any]]:
    """Converts a list of country long names into a list of country short names.

    Parameters
    ----------
    countries : list
        Country long names.

    Returns
    -------
    dict
        Mapping used for the conversion.
    list
        Country short names.
    """
    mapping = {v: k for k, v in COUNTRY_NAMES_MAPPING.items()}
    return mapping, [mapping[country] for country in countries]


def mapping_freq_names(freq: str) -> str:
    """Converts a short frequency name into a long frequency name.

    Parameters
    ----------
    freq : str
        Short frequency name.

    Returns
    -------
    str
        Long frequency name.
    """
    mapping = {
        "s": "seconds",
        "H": "hours",
        "D": "days",
        "W": "weeks",
        "M": "months",
        "Q": "quarters",
        "Y": "years",
    }
    return mapping[freq]



from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd
import streamlit as st


@st.cache()
def remove_empty_cols(df: pd.DataFrame) -> Tuple[pd.DataFrame, List[Any]]:
    """Remove columns with strictly less than 2 distinct values in input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be checked and potentially removed.

    Returns
    -------
    pd.DataFrame
        Dataframe with empty columns removed.
    list
        List of columns that have been removed.
    """
    count_cols = df.nunique(dropna=False)
    empty_cols = list(count_cols[count_cols < 2].index)
    return df.drop(empty_cols, axis=1), empty_cols


def print_empty_cols(empty_cols: List[Any]) -> None:
    """Displays a message in streamlit dashboard if the input list is not empty.

    Parameters
    ----------
    empty_cols : list
        List of columns that have been removed.
    """
    L = len(empty_cols)
    if L > 0:
        st.error(
            f'The following column{"s" if L > 1 else ""} ha{"ve" if L > 1 else "s"} been removed because '
            f'{"they have" if L > 1 else "it has"} <= 1 distinct values: {", ".join(empty_cols)}'
        )


@st.cache(suppress_st_warning=True)
def format_date_and_target(
    df_input: pd.DataFrame,
    date_col: str,
    target_col: str,
    config: Dict[Any, Any],
    load_options: Dict[Any, Any],
) -> pd.DataFrame:
    """Formats date and target columns of input dataframe.

    Parameters
    ----------
    df_input : pd.DataFrame
        Input dataframe whose columns will be formatted.
    date_col : str
        Name of date column in input dataframe.
    target_col : str
        Name of target column in input dataframe.
    config : Dict
        Lib configuration dictionary.
    load_options : Dict
        Loading options selected by user.

    Returns
    -------
    pd.DataFrame
        Dataframe with columns formatted.
    """
    df = df_input.copy()  # To avoid CachedObjectMutationWarning
    df = _format_date(df, date_col, load_options, config)
    df = _format_target(df, target_col, config)
    df = _rename_cols(df, date_col, target_col)
    return df


def _format_date(
    df: pd.DataFrame, date_col: str, load_options: Dict[Any, Any], config: Dict[Any, Any]
) -> pd.DataFrame:
    """Formats date column of input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be formatted.
    date_col : str
        Name of date column in input dataframe.
    load_options : Dict
        Loading options selected by user.
    config : Dict
        Lib config dictionary containing information about default date format.

    Returns
    -------
    pd.DataFrame
        Dataframe with date column formatted.
    """
    try:
        date_series = pd.to_datetime(df[date_col])
        if __check_date_format(date_series) | (
            config["dataprep"]["date_format"] != load_options["date_format"]
        ):
            date_series = pd.to_datetime(df[date_col], format=load_options["date_format"])
        df[date_col] = date_series
        days_range = (df[date_col].max() - df[date_col].min()).days
        sec_range = (df[date_col].max() - df[date_col].min()).seconds
        if ((days_range < 1) & (sec_range < 1)) | (np.isnan(days_range) & np.isnan(sec_range)):
            st.error(
                "Please select the correct date column (selected column has a time range < 1s)."
            )
            st.stop()
        return df
    except:
        st.error(
            "Please select a valid date format (selected column can't be converted into date)."
        )
        st.stop()


def __check_date_format(date_series: pd.Series) -> bool:
    """Checks whether the date column has been correctly converted to datetime.

    Parameters
    ----------
    date_series : pd.Series
        Date column that has been converted.

    Returns
    -------
    bool
        False if conversion has not worked correctly, True otherwise.
    """
    test1 = date_series.map(lambda x: x.year).nunique() < 2
    test2 = date_series.map(lambda x: x.month).nunique() < 2
    test3 = date_series.map(lambda x: x.day).nunique() < 2
    if test1 & test2 & test3:
        return True
    else:
        return False


def _format_target(df: pd.DataFrame, target_col: str, config: Dict[Any, Any]) -> pd.DataFrame:
    """Formats target column of input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be formatted.
    target_col : str
        Name of target column in input dataframe.

    Returns
    -------
    pd.DataFrame
        Dataframe with date column formatted.
    """
    try:
        df[target_col] = df[target_col].astype("float")
        if df[target_col].nunique() < config["validity"]["min_target_cardinality"]:
            st.error(
                "Please select the correct target column (should be numerical, not categorical)."
            )
            st.stop()
        return df
    except:
        st.error("Please select the correct target column (should be of type int or float).")
        st.stop()


def _rename_cols(df: pd.DataFrame, date_col: str, target_col: str) -> pd.DataFrame:
    """Renames date and target columns of input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be renamed.
    date_col : str
        Name of date column in input dataframe.
    target_col : str
        Name of target column in input dataframe.

    Returns
    -------
    pd.DataFrame
        Dataframe with columns renamed.
    """
    if (target_col != "y") and ("y" in df.columns):
        df = df.rename(columns={"y": "y_2"})
    if (date_col != "ds") and ("ds" in df.columns):
        df = df.rename(columns={"ds": "ds_2"})
    df = df.rename(columns={date_col: "ds", target_col: "y"})
    return df


# NB: date_col and target_col not used, only added to avoid unexpected caching when their values change
@st.cache()
def filter_and_aggregate_df(
    df_input: pd.DataFrame,
    dimensions: Dict[Any, Any],
    config: Dict[Any, Any],
    date_col: str,
    target_col: str,
) -> Tuple[pd.DataFrame, List[Any]]:
    """Filters and aggregates input dataframe according to dimensions dictionary specifications.

    Parameters
    ----------
    df_input : pd.DataFrame
        Input dataframe that will be filtered and/or aggregated.
    dimensions : Dict
        Filtering and aggregation specifications.
    config : Dict
        Lib configuration dictionary.
    date_col : str
        Name of date column in input dataframe.
    target_col : str
        Name of target column in input dataframe.

    Returns
    -------
    pd.DataFrame
        Dataframe filtered and/or aggregated.
    list
        List of columns removed from input dataframe.
    """
    df = df_input.copy()  # To avoid CachedObjectMutationWarning
    df = _filter(df, dimensions)
    df, cols_to_drop = _format_regressors(df, config)
    df = _aggregate(df, dimensions)
    return df, cols_to_drop


def _filter(df: pd.DataFrame, dimensions: Dict[Any, Any]) -> pd.DataFrame:
    """Filters input dataframe according to dimensions dictionary specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be filtered and/or aggregated.
    dimensions : Dict
        Filtering specifications.

    Returns
    -------
    pd.DataFrame
        Filtered dataframe.
    """
    filter_cols = list(set(dimensions.keys()) - {"agg"})
    for col in filter_cols:
        df = df.loc[df[col].isin(dimensions[col])]
    return df.drop(filter_cols, axis=1)


def _format_regressors(df: pd.DataFrame, config: Dict[Any, Any]) -> Tuple[pd.DataFrame, List[Any]]:
    """Format some columns in input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be formatted.
    config : Dict
        Lib configuration dictionary.

    Returns
    -------
    pd.DataFrame
        Formatted dataframe.
    list
        List of columns removed from input dataframe.
    """
    cols_to_drop = []
    for col in set(df.columns) - {"ds", "y"}:
        if df[col].nunique(dropna=False) < 2:
            cols_to_drop.append(col)
        elif df[col].nunique(dropna=False) == 2:
            df[col] = df[col].map(dict(zip(df[col].unique(), [0, 1])))
        elif df[col].nunique() <= config["validity"]["max_cat_reg_cardinality"]:
            df = __one_hot_encoding(df, col)
        else:
            try:
                df[col] = df[col].astype("float")
            except:
                cols_to_drop.append(col)
    return df.drop(cols_to_drop, axis=1), cols_to_drop


def __one_hot_encoding(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """Applies one-hot encoding to some columns of input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe whose columns will be one-hot encoded.
    col : list
        List of columns to one-hot encode.

    Returns
    -------
    pd.DataFrame
        One-hot encoded dataframe.
    """
    df = pd.concat([df, pd.get_dummies(df[col], prefix=col)], axis=1)
    return df.drop(col, axis=1)


def print_removed_cols(cols_removed: List[Any]) -> None:
    """Displays a message in streamlit dashboard if the input list is not empty.

    Parameters
    ----------
    cols_removed : list
        List of columns that have been removed.
    """
    L = len(cols_removed)
    if L > 0:
        st.error(
            f'The following column{"s" if L > 1 else ""} ha{"ve" if L > 1 else "s"} been removed because '
            f'{"they are" if L > 1 else "it is"} neither the target, '
            f'nor a dimension, nor a potential regressor: {", ".join(cols_removed)}'
        )


def _aggregate(df: pd.DataFrame, dimensions: Dict[Any, Any]) -> pd.DataFrame:
    """Aggregates input dataframe according to dimensions dictionary specifications.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe that will be filtered and/or aggregated.
    dimensions : Dict
        Filtering specifications.

    Returns
    -------
    pd.DataFrame
        Aggregated dataframe.
    """
    cols_to_agg = set(df.columns) - {"ds", "y"}
    agg_dict = {col: "mean" if df[col].nunique() > 2 else "max" for col in cols_to_agg}
    agg_dict["y"] = dimensions["agg"].lower()
    return df.groupby("ds").agg(agg_dict).reset_index()


@st.cache()
def format_datetime(df_input: pd.DataFrame, resampling: Dict[Any, Any]) -> pd.DataFrame:
    """Formats date column to datetime in input dataframe.

    Parameters
    ----------
    df_input : pd.DataFrame
        Input dataframe whose date column will be formatted to datetime.
    resampling : Dict
        Dictionary whose "freq" key contains the frequency of input dataframe.

    Returns
    -------
    pd.DataFrame
        Dataframe with date column formatted to datetime.
    """
    df = df_input.copy()  # To avoid CachedObjectMutationWarning
    if resampling["freq"][-1] in ["H", "s"]:
        df["ds"] = df["ds"].map(lambda x: x.strftime("%Y-%m-%d %H:%M:%S"))
        df["ds"] = pd.to_datetime(df["ds"])
    return df


@st.cache()
def resample_df(df_input: pd.DataFrame, resampling: Dict[Any, Any]) -> pd.DataFrame:
    """Resamples input dataframe according to resampling dictionary specifications.

    Parameters
    ----------
    df_input : pd.DataFrame
        Input dataframe that will be resampled.
    resampling : Dict
        Resampling specifications.

    Returns
    -------
    pd.DataFrame
        Resampled dataframe.
    """
    df = df_input.copy()  # To avoid CachedObjectMutationWarning
    if resampling["resample"]:
        cols_to_agg = set(df.columns) - {"ds", "y"}
        agg_dict = {col: "mean" if df[col].nunique() > 2 else "max" for col in cols_to_agg}
        agg_dict["y"] = resampling["agg"].lower()
        df = df.set_index("ds").resample(resampling["freq"][-1]).agg(agg_dict).reset_index()
    return df


def check_dataset_size(df: pd.DataFrame, config: Dict[Any, Any]) -> None:
    """Displays a message in streamlit dashboard and stops it if the input dataframe has not enough rows.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    config : Dict
        Lib configuration dictionary where the minimum number of rows is given.
    """
    if (
        len(df)
        <= config["validity"]["min_data_points_train"] + config["validity"]["min_data_points_val"]
    ):
        st.error(
            f"The dataset has not enough data points ({len(df)} data points only) to make a forecast. "
            f"Please resample with a higher frequency or change cleaning options."
        )
        st.stop()


def check_future_regressors_df(
    datasets: Dict[Any, Any],
    dates: Dict[Any, Any],
    params: Dict[Any, Any],
    resampling: Dict[Any, Any],
    date_col: str,
    dimensions: Dict[Any, Any],
) -> bool:
    """Displays a message if the future regressors dataframe is incorrect and says whether or not to use it afterwards.

    Parameters
    ----------
    datasets : Dict
        Dictionary storing all dataframes.
    dates : Dict
        Dictionary containing future forecasting dates information.
    params : Dict
        Dictionary containing all model parameters and list of selected regressors.
    resampling : Dict
        Dictionary containing dataset frequency information.
    date_col : str
        Name of date column.
    dimensions : Dict
        Dictionary containing dimensions information.

    Returns
    -------
    bool
        Whether or not to use regressors for future forecast.
    """
    use_regressors = False
    if "future_regressors" in datasets.keys():
        # Check date column
        if date_col not in datasets["future_regressors"].columns:
            st.error(
                f"Date column '{date_col}' not found in the dataset provided for future regressors."
            )
            st.stop()
        # Check number of distinct dates
        N_dates_input = datasets["future_regressors"][date_col].nunique()
        N_dates_expected = len(
            pd.date_range(
                start=dates["forecast_start_date"],
                end=dates["forecast_end_date"],
                freq=resampling["freq"],
            )
        )
        if N_dates_input != N_dates_expected:
            st.error(
                f"The dataset provided for future regressors has the right number of distinct dates "
                f"(expected {N_dates_expected}, found {N_dates_input}). "
                f"Please make sure that the date column goes from {dates['forecast_start_date'].strftime('%Y-%m-%d')} "
                f"to {dates['forecast_end_date'].strftime('%Y-%m-%d')} at frequency {resampling['freq']} "
                f"without skipping any date in this range."
            )
            st.stop()
        # Check regressors
        regressors_expected = set(params["regressors"].keys())
        input_cols = set(datasets["future_regressors"])
        if len(input_cols.intersection(regressors_expected)) != len(regressors_expected):
            missing_regressors = [reg for reg in regressors_expected if reg not in input_cols]
            if len(missing_regressors) > 1:
                st.error(
                    f"Columns {', '.join(missing_regressors[:-1])} and {missing_regressors[-1]} are missing "
                    f"in the dataset provided for future regressors."
                )
            else:
                st.error(
                    f"Column {missing_regressors[0]} is missing in the dataset provided for future regressors."
                )
            st.stop()
        # Check dimensions
        dim_expected = {dim for dim in dimensions.keys() if dim != "agg"}
        if len(input_cols.intersection(dim_expected)) != len(dim_expected):
            missing_dim = [dim for dim in dim_expected if dim not in input_cols]
            if len(missing_dim) > 1:
                st.error(
                    f"Dimension columns {', '.join(missing_dim[:-1])} and {missing_dim[-1]} are missing "
                    f"in the dataset provided for future regressors."
                )
            else:
                st.error(
                    f"Dimension column {missing_dim[0]} is missing in the dataset provided for future regressors."
                )
            st.stop()
        use_regressors = True
    return use_regressors


def prepare_future_df(
    datasets: Dict[Any, Any],
    dates: Dict[Any, Any],
    date_col: str,
    target_col: str,
    dimensions: Dict[Any, Any],
    load_options: Dict[Any, Any],
    config: Dict[Any, Any],
    resampling: Dict[Any, Any],
) -> Tuple[pd.DataFrame, Dict[Any, Any]]:
    """Applies data preparation to the dataset provided with future regressors.

    Parameters
    ----------
    datasets : Dict
        Dictionary storing all dataframes.
    dates : Dict
        Dictionary containing future forecasting dates information.
    date_col : str
        Name of date column.
    target_col : str
        Name of target column.
    dimensions : Dict
        Dictionary containing dimensions information.
    load_options : Dict
        Loading options selected by user.
    config : Dict
        Lib configuration dictionary.
    resampling : Dict
        Resampling specifications.

    Returns
    -------
    pd.DataFrame
        Prepared  future dataframe.
    dict
        Dictionary storing all dataframes.
    """
    if "future_regressors" in datasets.keys():
        future = datasets["future_regressors"]
        future[target_col] = 0
        future = pd.concat([datasets["uploaded"][list(future.columns)], future], axis=0)
        future, _ = remove_empty_cols(future)
        future = format_date_and_target(future, date_col, target_col, config, load_options)
        future, _ = filter_and_aggregate_df(future, dimensions, config, date_col, target_col)
        future = format_datetime(future, resampling)
        future = resample_df(future, resampling)
        datasets["full"] = future.loc[future["ds"] < dates["forecast_start_date"]]
        future = future.drop("y", axis=1)
    else:
        future_dates = pd.date_range(
            start=datasets["full"].ds.min(),
            end=dates["forecast_end_date"],
            freq=dates["forecast_freq"],
        )
        future = pd.DataFrame(future_dates, columns=["ds"])
    return future, datasets


@st.cache()
def add_cap_and_floor_cols(df_input: pd.DataFrame, params: Dict[Any, Any]) -> pd.DataFrame:
    """Resamples input dataframe according to resampling dictionary specifications.

    Parameters
    ----------
    df_input : pd.DataFrame
        Input dataframe that will be resampled.
    params : Dict
        Model parameters.

    Returns
    -------
    pd.DataFrame
        Dataframe with cap and floor columns if specified.
    """
    df = df_input.copy()  # To avoid CachedObjectMutationWarning
    if params["other"]["growth"] == "logistic":
        df["cap"] = params["saturation"]["cap"]
        df["floor"] = params["saturation"]["floor"]
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['streamlit_prophet',
 'streamlit_prophet.app',
 'streamlit_prophet.cli',
 'streamlit_prophet.lib',
 'streamlit_prophet.lib.dataprep',
 'streamlit_prophet.lib.evaluation',
 'streamlit_prophet.lib.exposition',
 'streamlit_prophet.lib.inputs',
 'streamlit_prophet.lib.models',
 'streamlit_prophet.lib.utils']

package_data = \
{'': ['*'],
 'streamlit_prophet': ['config/config_instructions.toml',
                       'config/config_instructions.toml',
                       'config/config_instructions.toml',
                       'config/config_readme.toml',
                       'config/config_readme.toml',
                       'config/config_readme.toml',
                       'config/config_streamlit.toml',
                       'config/config_streamlit.toml',
                       'config/config_streamlit.toml',
                       'references/input_format.png',
                       'references/input_format.png',
                       'references/logo.png',
                       'references/logo.png',
                       'report/config/.gitignore',
                       'report/data/.gitignore',
                       'report/plots/.gitignore']}

install_requires = \
['fbprophet==0.7.1',
 'holidays>=0.11.1,<0.12.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.1.5,<2.0.0',
 'plotly>=4.11.0,<5.0.0',
 'pystan==2.19.1.1',
 'rich>=10.1.0,<11.0.0',
 'scipy>=1.6.3,<2.0.0',
 'streamlit==0.80.0',
 'typer[all]>=0.3.2,<0.4.0',
 'vacances-scolaires-france>=0.8.0,<0.9.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['streamlit_prophet = streamlit_prophet.cli.__main__:app']}

setup_kwargs = {
    'name': 'streamlit-prophet',
    'version': '1.0.0',
    'description': 'Deploy a Streamlit app to train, evaluate and optimize a Prophet forecasting model visually.',
    'long_description': '<div align="center">\n\n![](streamlit_prophet/references/logo.png)\n\n[![CI status](https://github.com/artefactory-global/streamlit_prophet/actions/workflows/ci.yml/badge.svg?branch%3Amain&event%3Apush)](https://github.com/artefactory-global/streamlit_prophet/actions/workflows/ci.yml?query=branch%3Amain)\n[![Python Version](https://img.shields.io/badge/python-3.7%20%7C%203.8%20%7C%203.9-blue.svg)](#supported-python-versions)\n[![Dependencies Status](https://img.shields.io/badge/dependabots-active-informational.svg)](https://github.com/artefactory-global/streamlit_prophet/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-informational.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-informational?logo=pre-commit&logoColor=white)](https://github.com/artefactory-global/streamlit_prophet/blob/main/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%F0%9F%9A%80-semantic%20versions-informational.svg)](https://github.com/artefactory-global/streamlit_prophet/releases)\n[![License](https://img.shields.io/badge/License-MIT-informational.svg)](https://github.com/artefactory-global/streamlit_prophet/blob/main/LICENSE)\n\nDeploy a [Streamlit](https://streamlit.io/) app to train, evaluate and optimize a [Prophet](https://facebook.github.io/prophet/) forecasting model visually\n\n</div>\n\nhttps://user-images.githubusercontent.com/56996548/126762714-f2d3f3a1-7098-4a86-8c60-0a69d0f913a7.mp4\n\n## 💻 Requirements\n\n### Python version\n* Main supported version : <strong>3.7</strong> <br>\n* Other supported versions : <strong>3.8</strong> & <strong>3.9</strong>\n\nPlease make sure you have one of these versions installed to be able to run the app on your machine.\n\n### Operating System\nWindows users have to install [WSL2](https://docs.microsoft.com/en-us/windows/wsl/) to download the package. \nThis is due to an incompatibility between Windows and Prophet\'s main dependency (pystan). \nOther operating systems should work fine.\n\n## ⚙️ Installation\n\n\n### Create a virtual environment (optional)\nWe strongly advise to create and activate a new virtual environment, to avoid any dependency issue.\n\nFor example with conda:\n```bash\npip install conda; conda create -n streamlit_prophet python=3.7; conda activate streamlit_prophet\n```\n\nOr with virtualenv:\n```bash\npip install virtualenv; python3.7 -m virtualenv streamlit_prophet --python=python3.7; source streamlit_prophet/bin/activate\n```\n\n\n### Install package\nInstall the package from PyPi (it should take a few minutes):\n```bash\npip install -U streamlit_prophet\n```\n\nOr from the main branch of this repository:\n```bash\npip install git+https://github.com/artefactory-global/streamlit_prophet.git@main\n```\n\n\n## 📈 Usage\n\nOnce installed, run the following command from CLI to open the app in your default web browser:\n\n```bash\nstreamlit_prophet deploy dashboard\n```\n\nNow you can train, evaluate and optimize forecasting models in a few clicks.\nAll you have to do is to upload a time series dataset. \nThis dataset should be a csv file that contains a date column, a target column and optionally some features, like on the example below:\n\n![](streamlit_prophet/references/input_format.png)\n\nThen, follow the guidelines in the sidebar to:\n\n* <strong>Prepare data</strong>: Filter, aggregate, resample and/or clean your dataset.\n* <strong>Choose model parameters</strong>: Default parameters are available but you can tune them.\nLook at the tooltips to understand how each parameter is impacting forecasts.\n* <strong>Select evaluation method</strong>: Define the evaluation process, the metrics and the granularity to\nassess your model performance.\n* <strong>Make a forecast</strong>: Make a forecast on future dates that are not included in your dataset,\nwith the model previously trained.\n\nOnce you are satisfied, click on "save experiment" to download all plots and data locally.\n\n\n## 🛠️ How to contribute ?\n\nAll contributions, ideas and bug reports are welcome! \nWe encourage you to open an [issue](https://github.com/artefactory-global/streamlit_prophet/issues) for any change you would like to make on this project.\n\n\nFor more information, see [`CONTRIBUTING`](https://github.com/artefactory-global/streamlit_prophet/blob/main/CONTRIBUTING.md) instructions.\nIf you wish to containerize the app, see [`DOCKER`](https://github.com/artefactory-global/streamlit_prophet/blob/main/DOCKER.md) instructions.\n',
    'author': 'Maxime Lutel',
    'author_email': 'maxime.lutel@artefact.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/artefactory-global/streamlit_prophet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from typing import Any, Dict, Tuple

import datetime
from datetime import timedelta

import pandas as pd
import streamlit as st
from streamlit_prophet.lib.dataprep.split import (
    get_cv_cutoffs,
    get_max_possible_cv_horizon,
    get_train_end_date_default_value,
    print_cv_folds_dates,
    print_forecast_dates,
    raise_error_cv_dates,
)
from streamlit_prophet.lib.utils.mapping import (
    convert_into_nb_of_days,
    convert_into_nb_of_seconds,
    mapping_freq_names,
)


def input_train_dates(
    df: pd.DataFrame,
    use_cv: bool,
    config: Dict[Any, Any],
    resampling: Dict[Any, Any],
    dates: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Lets the user enter training dates.

    Parameters
    ----------
    df : Dict
        Prepared dataset (after filtering, resampling, cleaning).
    use_cv : bool
        Whether or not cross-validation is used.
    config : Dict
        Lib config dictionary containing information needed to set default dates displayed in streamlit.
    resampling : Dict
        Dictionary containing dataset frequency information.
    dates : Dict
        Empty dictionary.

    Returns
    -------
    dict
        Dictionary containing training dates information.
    """
    col1, col2 = st.beta_columns(2)
    set_name = "CV" if use_cv else "Training"
    dates["train_start_date"] = col1.date_input(
        f"{set_name} start date", value=df.ds.min(), min_value=df.ds.min(), max_value=df.ds.max()
    )
    default_end_date = get_train_end_date_default_value(df, dates, resampling, config, use_cv)
    dates["train_end_date"] = col2.date_input(
        f"{set_name} end date",
        value=default_end_date,
        min_value=dates["train_start_date"] + timedelta(days=1),
        max_value=df.ds.max(),
    )
    return dates


def input_val_dates(
    df: pd.DataFrame, dates: Dict[Any, Any], config: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Lets the user enter validation dates.

    Parameters
    ----------
    df : Dict
        Prepared dataset (after filtering, resampling, cleaning).
    dates : Dict
        Dictionary containing training dates information.
    config : Dict
        Lib config dictionary containing information needed to set default dates displayed in streamlit.

    Returns
    -------
    dict
        Dictionary containing training and validation dates information.
    """
    col1, col2 = st.beta_columns(2)
    dates["val_start_date"] = col1.date_input(
        "Validation start date",
        value=dates["train_end_date"] + timedelta(days=config["split"]["gap_train_valid"]),
        min_value=dates["train_end_date"] + timedelta(days=1),
        max_value=df.ds.max(),
    )
    dates["val_end_date"] = col2.date_input(
        "Validation end date",
        value=df.ds.max(),
        min_value=dates["val_start_date"] + timedelta(days=1),
        max_value=df.ds.max(),
    )
    return dates


def input_cv(
    dates: Dict[Any, Any],
    resampling: Dict[Any, Any],
    config: Dict[Any, Any],
    readme: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Lets the user enter cross-validation specifications.

    Parameters
    ----------
    dates : Dict
        Dictionary containing training dates information.
    resampling : Dict
        Dictionary containing dataset frequency information.
    config : Dict
        Lib config dictionary containing information needed to set default dates displayed in streamlit.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Dictionary containing training dates and cross-validation specifications.
    """
    dates["n_folds"] = st.number_input(
        "Number of CV folds",
        min_value=1,
        value=config["split"]["CV"],
        help=readme["tooltips"]["cv_n_folds"],
    )
    freq = resampling["freq"][-1]
    max_possible_horizon = get_max_possible_cv_horizon(dates, resampling)
    dates["folds_horizon"] = st.number_input(
        f"Horizon of each fold (in {mapping_freq_names(freq)})",
        min_value=3,
        max_value=max_possible_horizon,
        value=min(config["horizon"][freq], max_possible_horizon),
        help=readme["tooltips"]["cv_horizon"],
    )
    dates["cutoffs"] = get_cv_cutoffs(dates, freq)
    print_cv_folds_dates(dates, freq)
    raise_error_cv_dates(dates, resampling, config)
    return dates


def input_forecast_dates(
    df: pd.DataFrame,
    dates: Dict[Any, Any],
    resampling: Dict[Any, Any],
    config: Dict[Any, Any],
    readme: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Lets the user enter future forecast dates.

    Parameters
    ----------
    df : pd.DataFrame
        Prepared dataset (after filtering, resampling, cleaning).
    dates : Dict
        Dictionary containing dates information.
    resampling : Dict
        Dictionary containing dataset frequency information.
    config : Dict
        Lib config dictionary containing information needed to set default dates displayed in streamlit.
    readme : Dict
        Dictionary containing tooltips to guide user's choices.

    Returns
    -------
    dict
        Dictionary containing future forecast dates information.
    """
    forecast_freq_name = mapping_freq_names(resampling["freq"][-1])
    dates["forecast_horizon"] = st.number_input(
        f"Forecast horizon in {forecast_freq_name}",
        min_value=1,
        value=config["horizon"][resampling["freq"][-1]],
        help=readme["tooltips"]["forecast_horizon"],
    )
    if forecast_freq_name in ["seconds", "hours"]:
        dates["forecast_start_date"] = df.ds.max() + timedelta(seconds=1)
        timedelta_horizon = convert_into_nb_of_seconds(
            resampling["freq"][-1], dates["forecast_horizon"]
        )
        dates["forecast_end_date"] = dates["forecast_start_date"] + timedelta(
            seconds=timedelta_horizon
        )
    else:
        dates["forecast_start_date"] = df.ds.max() + timedelta(days=1)
        timedelta_horizon = convert_into_nb_of_days(
            resampling["freq"][-1], dates["forecast_horizon"]
        )
        dates["forecast_end_date"] = dates["forecast_start_date"] + timedelta(
            days=timedelta_horizon
        )
    dates["forecast_freq"] = str(resampling["freq"])
    print_forecast_dates(dates, resampling)
    return dates


def input_waterfall_dates(
    forecast_df: pd.DataFrame, resampling: Dict[Any, Any]
) -> Tuple[datetime.date, datetime.date]:
    """Lets the user enter dates for waterfall components chart.

    Parameters
    ----------
    forecast_df : pd.DataFrame
        Dataframe with forecast and components.
    resampling : Dict
        Resampling specifications (granularity, dataset frequency).

    Returns
    -------
    datetime.date
        Waterfall start date.
    datetime.date
        Waterfall end date.
    """
    max_date = forecast_df.loc[~pd.isnull(forecast_df["trend"])]["ds"].max()
    col1, col2 = st.beta_columns(2)
    start_date = col1.date_input(
        "Start date", value=forecast_df.ds.min(), min_value=forecast_df.ds.min(), max_value=max_date
    )
    freq = resampling["freq"][-1]
    n_periods = col2.number_input(
        f"Number of {mapping_freq_names(freq)} to focus on", value=1, min_value=1
    )
    end_date = start_date + timedelta(days=convert_into_nb_of_days(freq, n_periods))
    return start_date, end_date
